#!/usr/bin/env python3
#
# SPDX-License-Identifier: MIT
#
# This utility clones the openrazer repository, extracts the known VID/PID combos
# for "RazerBlade" keyboards and spits out a libinput-compatible quirks file
# with those keyboards listed as "Internal".
#
# Usage:
# $ cd path/to/libinput.git/
# $ ./tools/razer-quirks-list.py

from pathlib import Path
from collections import namedtuple
import tempfile
import shutil
import subprocess
import sys


KeyboardDescriptor = namedtuple("KeyboardDescriptor", ["vid", "pid", "name"])

with tempfile.TemporaryDirectory() as tmpdir:
    # Clone the openrazer directory and add its Python modyle basedir
    # to our sys.path so we can import it
    subprocess.run(
        [
            "git",
            "clone",
            "--quiet",
            "--depth=1",
            "https://github.com/openrazer/openrazer",
        ],
        cwd=tmpdir,
        check=True,
    )
    sys.path.append(str(Path(tmpdir) / "openrazer" / "daemon"))

    import openrazer_daemon.hardware.keyboards as razer_keyboards  # type: ignore

    keyboards: list[KeyboardDescriptor] = []

    # All classnames for the keyboards we care about start with RazerBlade
    for clsname in filter(lambda c: c.startswith("RazerBlade"), dir(razer_keyboards)):
        kbd = getattr(razer_keyboards, clsname)
        try:
            k = KeyboardDescriptor(name=clsname, vid=kbd.USB_VID, pid=kbd.USB_PID)
            keyboards.append(k)
        except AttributeError:
            continue

    tmpfile = Path(tmpdir) / "30-vendor-razer.quirks"
    try:
        quirksfile = next(Path("quirks").glob("30-vendor-razer.quirks"))
    except StopIteration:
        print("Unable to find the razer quirks file.")
        raise SystemExit(1)

    with open(tmpfile, "w") as output:
        with open(quirksfile) as input:
            for line in input:
                output.write(line)
                if line.startswith("# AUTOGENERATED"):
                    output.write("\n")
                    break

        for kbd in sorted(keyboards, key=lambda k: k.vid << 16 | k.pid):
            print(
                "\n".join(
                    (
                        f"[{kbd.name} Keyboard]",
                        "MatchUdevType=keyboard",
                        "MatchBus=usb",
                        f"MatchVendor=0x{kbd.vid:04X}",
                        f"MatchProduct=0x{kbd.pid:04X}",
                        "AttrKeyboardIntegration=internal",
                    )
                ),
                file=output,
            )
            print(file=output)

    shutil.copy(tmpfile, quirksfile)



#!/usr/bin/env python3
# -*- coding: utf-8
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2020 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the 'Software'),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#
#
# Prints the down/up state of each touch slot
#
# Input is a libinput record yaml file

import argparse
import enum
import sys
import yaml
import libevdev


class Slot:
    class State(enum.Enum):
        NONE = "NONE"
        BEGIN = "BEGIN"
        UPDATE = "UPDATE"
        END = "END"

    def __init__(self, index):
        self._state = Slot.State.NONE
        self.index = index
        self.used = False

    def begin(self):
        assert self.state == Slot.State.NONE
        self.state = Slot.State.BEGIN

    def end(self):
        assert self.state in (Slot.State.BEGIN, Slot.State.UPDATE)
        self.state = Slot.State.END

    def sync(self):
        if self.state == Slot.State.BEGIN:
            self.state = Slot.State.UPDATE
        elif self.state == Slot.State.END:
            self.state = Slot.State.NONE

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, newstate):
        assert newstate in Slot.State

        if newstate != Slot.State.NONE:
            self.used = True
        self._state = newstate

    @property
    def is_active(self):
        return self.state in (Slot.State.BEGIN, Slot.State.UPDATE)

    def __str__(self):
        return "+" if self.state in (Slot.State.BEGIN, Slot.State.UPDATE) else " "


def main(argv):
    parser = argparse.ArgumentParser(description="Print the state of touches over time")
    parser.add_argument(
        "--use-st", action="store_true", help="Ignore slots, use the BTN_TOOL bits"
    )
    parser.add_argument(
        "path", metavar="recording", nargs=1, help="Path to libinput-record YAML file"
    )
    args = parser.parse_args()

    yml = yaml.safe_load(open(args.path[0]))
    device = yml["devices"][0]
    absinfo = device["evdev"]["absinfo"]
    try:
        nslots = absinfo[libevdev.EV_ABS.ABS_MT_SLOT.value][1] + 1
    except KeyError:
        args.use_st = True

    tool_slot_map = {
        libevdev.EV_KEY.BTN_TOOL_FINGER: 0,
        libevdev.EV_KEY.BTN_TOOL_PEN: 0,
        libevdev.EV_KEY.BTN_TOOL_DOUBLETAP: 1,
        libevdev.EV_KEY.BTN_TOOL_TRIPLETAP: 2,
        libevdev.EV_KEY.BTN_TOOL_QUADTAP: 3,
        libevdev.EV_KEY.BTN_TOOL_QUINTTAP: 4,
    }
    if args.use_st:
        for bit in tool_slot_map:
            if bit.value in device["evdev"]["codes"][libevdev.EV_KEY.value]:
                nslots = max(nslots, tool_slot_map[bit])

    slots = [Slot(i) for i in range(0, nslots)]
    # We claim the first slots are used just to make the formatting
    # more consistent
    for i in range(min(5, len(slots))):
        slots[i].used = True

    slot = 0
    last_time = None
    last_slot_state = None
    header = "Timestamp | Rel time |     Slots     |"
    print(header)
    print("-" * len(header))

    def events():
        for event in device["events"]:
            for evdev in event["evdev"]:
                yield evdev

    for evdev in events():
        e = libevdev.InputEvent(
            code=libevdev.evbit(evdev[2], evdev[3]),
            value=evdev[4],
            sec=evdev[0],
            usec=evdev[1],
        )

        # single-touch formatting is simpler than multitouch, it'll just
        # show the highest finger down rather than the correct output.
        if args.use_st:
            if e.code in tool_slot_map:
                slot = tool_slot_map[e.code]
                s = slots[slot]
                if e.value:
                    s.begin()
                else:
                    s.end()
        else:
            if e.code == libevdev.EV_ABS.ABS_MT_SLOT:
                slot = e.value
                s = slots[slot]
                # bcm5974 cycles through slot numbers, so let's say all below
                # our current slot number was used
                for sl in slots[: slot + 1]:
                    sl.used = True
            else:
                s = slots[slot]
                if e.code == libevdev.EV_ABS.ABS_MT_TRACKING_ID:
                    if e.value == -1:
                        s.end()
                    else:
                        s.begin()
                elif e.code in (
                    libevdev.EV_ABS.ABS_MT_POSITION_X,
                    libevdev.EV_ABS.ABS_MT_POSITION_Y,
                    libevdev.EV_ABS.ABS_MT_PRESSURE,
                    libevdev.EV_ABS.ABS_MT_TOUCH_MAJOR,
                    libevdev.EV_ABS.ABS_MT_TOUCH_MINOR,
                ):
                    # If recording started after touch down
                    if s.state == Slot.State.NONE:
                        s.begin()

        if e.code == libevdev.EV_SYN.SYN_REPORT:
            current_slot_state = tuple(s.is_active for s in slots)

            if current_slot_state != last_slot_state:
                if last_time is None:
                    last_time = e.sec * 1000000 + e.usec
                    tdelta = 0
                else:
                    t = e.sec * 1000000 + e.usec
                    tdelta = int((t - last_time) / 1000) / 1000
                    last_time = t

                fmt = " | ".join([str(s) for s in slots if s.used])
                print(
                    "{:2d}.{:06d} | {:+7.3f}s | {}".format(e.sec, e.usec, tdelta, fmt)
                )

                last_slot_state = current_slot_state

            for s in slots:
                s.sync()


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the 'Software'),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import sys

try:
    import pyudev
except ModuleNotFoundError as e:
    print("Error: {}".format(str(e)), file=sys.stderr)
    print(
        "One or more python modules are missing. Please install those "
        "modules and re-run this tool."
    )
    sys.exit(1)


def list_devices():
    devices = {}
    context = pyudev.Context()
    for device in context.list_devices(subsystem="input"):
        if (device.device_node or "").startswith("/dev/input/event"):
            parent = device.parent
            if parent is not None:
                name = parent.properties["NAME"] or ""
                # The udev name includes enclosing quotes
                devices[device.device_node] = name[1:-1]

    def versionsort(key):
        return int(key[len("/dev/input/event") :])

    for k in sorted(devices, key=versionsort):
        print(f"{k}:\t{devices[k]}")


class HidDevice:
    def __init__(self, name, driver, vendor, product, devpath):
        self.name = name
        self.driver = driver
        self.vendor = vendor
        self.product = product
        self.devpath = devpath
        self.hidraws = []
        self.evdevs = []


def list_hid_devices():
    devices = []
    context = pyudev.Context()
    for device in context.list_devices(subsystem="hid"):
        name = device.properties.get("HID_NAME")
        driver = device.properties.get("DRIVER")
        devpath = device.properties.get("DEVPATH")
        id = device.properties.get("HID_ID") or "0:0:0"
        _, vendor, product = (int(x, 16) for x in id.split(":"))
        devices.append(HidDevice(name, driver, vendor, product, devpath))

    for device in context.list_devices(subsystem="hidraw"):
        devpath = device.properties["DEVPATH"]

        for hid in devices:
            if devpath.startswith(hid.devpath):
                hid.hidraws.append(f"'{device.device_node}'")

    for device in context.list_devices(subsystem="input"):
        if (device.device_node or "").startswith("/dev/input/event"):
            devpath = device.properties["DEVPATH"]

            for hid in devices:
                if devpath.startswith(hid.devpath):
                    hid.evdevs.append(f"'{device.device_node}'")

    print("hid:")
    for d in devices:
        print(f"- name:   '{d.name}'")
        print(f"  id:     '{d.vendor:04x}:{d.product:04x}'")
        print(f"  driver: '{d.driver}'")
        print(f"  hidraw: [{', '.join(h for h in d.hidraws)}]")
        print(f"  evdev:  [{', '.join(h for h in d.evdevs)}]")
        print("")


def main():
    parser = argparse.ArgumentParser(description="List kernel devices")
    parser.add_argument("--hid", action="store_true", default=False)
    args = parser.parse_args()

    if args.hid:
        list_hid_devices()
    else:
        list_devices()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("Exited on user request")



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

from pathlib import Path
from dataclasses import dataclass

import argparse
import itertools
import os
import sys


@dataclass
class WhitespaceError:
    message: str
    lineno: int
    nlines: int = 1
    column: None | int = None
    ncolumns: int = 1


def test_duplicate_empty_lines(lines: list[str]) -> list[WhitespaceError]:
    errors = []
    for idx, (l1, l2) in enumerate(itertools.pairwise(lines)):
        if not l1 and not l2:
            errors.append(WhitespaceError("Duplicated empty lines", idx, nlines=2))
    return errors


def test_tab_after_space(lines: list[str]) -> list[WhitespaceError]:
    errors = []
    for idx, l in enumerate(lines):
        index = l.find(" \t")
        if index > -1:
            errors.append(
                WhitespaceError(
                    "Tab after space", idx, nlines=index, column=index, ncolumns=2
                )
            )
    return errors


def test_trailing_whitespace(lines: list[str]) -> list[WhitespaceError]:
    errors = []
    for idx, l in enumerate(lines):
        if l.rstrip() != l:
            errors.append(WhitespaceError("Trailing whitespace", idx))
    return errors


def test_empty_line_between_braces(lines: list[str]) -> list[WhitespaceError]:
    errors = []
    for idx in range(len(lines) - 3):
        l1 = lines[idx]
        l2 = lines[idx + 1]
        l3 = lines[idx + 2]
        if l1.strip() == "}" and l3.strip() == "}" and l2.strip() == "":
            errors.append(WhitespaceError("Empty line between closing braces", idx + 1))
    return errors


def main():
    parser = argparse.ArgumentParser(description="Whitespace checker script")
    parser.add_argument(
        "files",
        metavar="FILES",
        type=Path,
        nargs="+",
        help="The files to check",
    )

    args = parser.parse_args()

    have_errors: bool = False

    if os.isatty(sys.stderr.fileno()):
        red = "\x1b[0;31m"
        reset = "\x1b[0m"
    else:
        red = ""
        reset = ""

    for file in args.files:
        lines = [l.rstrip("\n") for l in file.open().readlines()]

        errors = []
        errors.extend(test_tab_after_space(lines))
        errors.extend(test_trailing_whitespace(lines))
        if any(file.name.endswith(suffix) for suffix in [".c", ".h"]):
            if not file.parts[0] == "include":
                errors.extend(test_duplicate_empty_lines(lines))
                errors.extend(test_empty_line_between_braces(lines))

        for e in errors:
            print(f"{red}ERROR: {e.message} in {file}:{reset}", file=sys.stderr)
            print(f"{'-' * 72}", file=sys.stderr)
            lineno = max(0, e.lineno - 5)
            for idx, l in enumerate(lines[lineno : lineno + 10]):
                if e.lineno <= lineno + idx < e.lineno + e.nlines:
                    prefix = "->"
                    hl = red
                    nohl = reset
                else:
                    prefix = "  "
                    hl = ""
                    nohl = ""
                print(f"{hl}{lineno + idx:3d}: {prefix} {l.rstrip()}{nohl}")

            print(f"{'-' * 72}", file=sys.stderr)

        if errors:
            have_errors = True

    if have_errors:
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# This file is formatted with Python Black
#
# Run with pytest

from pathlib import Path

import configparser
import os
import pytest
import re

# see the IDs from
# https://github.com/torvalds/linux/blob/master/drivers/hid/hid-ids.h#L772
# https://github.com/torvalds/linux/blob/master/drivers/hid/hid-logitech-dj.c#L1826
logitech_receivers = [
    0xC50C,  # USB_DEVICE_ID_S510_RECEIVER
    0xC517,  # USB_DEVICE_ID_S510_RECEIVER_2
    0xC512,  # USB_DEVICE_ID_LOGITECH_CORDLESS_DESKTOP_LX500
    0xC513,  # USB_DEVICE_ID_MX3000_RECEIVER
    0xC51B,  # USB_DEVICE_ID_LOGITECH_27MHZ_MOUSE_RECEIVER
    0xC52B,  # USB_DEVICE_ID_LOGITECH_UNIFYING_RECEIVER
    0xC52F,  # USB_DEVICE_ID_LOGITECH_NANO_RECEIVER
    0xC532,  # USB_DEVICE_ID_LOGITECH_UNIFYING_RECEIVER_2
    0xC534,  # USB_DEVICE_ID_LOGITECH_NANO_RECEIVER_2
    0xC539,  # USB_DEVICE_ID_LOGITECH_NANO_RECEIVER_LIGHTSPEED_1
    0xC53F,  # USB_DEVICE_ID_LOGITECH_NANO_RECEIVER_LIGHTSPEED_1_1
    0xC53A,  # USB_DEVICE_ID_LOGITECH_NANO_RECEIVER_POWERPLAY
    0xC545,  # Bolt receiver, not listed in the kernel (yet)
    0xC547,  # Bolt receiver, not listed in the kernel (yet)
    0xC548,  # Bolt receiver, not listed in the kernel (yet)
]


def quirksdir():
    return Path(os.getenv("MESON_SOURCE_ROOT") or ".") / "quirks"


def pytest_generate_tests(metafunc):
    # for any function that takes a "quirksfile" argument return the path to
    # a quirks file
    if "quirksfile" in metafunc.fixturenames:
        metafunc.parametrize("quirksfile", [f for f in quirksdir().glob("*.quirks")])


def test_matches_are_valid(quirksfile):
    quirks = configparser.ConfigParser(strict=True)
    # Don't convert to lowercase
    quirks.optionxform = lambda option: option  # type: ignore
    quirks.read(quirksfile)

    for name, section in filter(lambda n: n != "DEFAULT", quirks.items()):
        bus = section.get("MatchBus")
        if bus is not None:
            assert bus in ("ps2", "usb", "bluetooth", "i2c", "spi")

        vid = section.get("MatchVendor")
        if vid is not None:
            assert re.match("0x[0-9A-F]{4}", vid), (
                f"{quirksfile}: {name}: {vid} must be uppercase hex (0xAB12)"
            )

        pid = section.get("MatchProduct")
        if pid is not None:
            assert re.match("0x[0-9A-F]{4}", pid), (
                f"{quirksfile}: {name}: {pid} must be uppercase hex (0xAB12)"
            )


def test_match_product_is_not_a_logitech_receiver(quirksfile):
    quirks = configparser.ConfigParser(strict=True)
    # Don't convert to lowercase
    quirks.optionxform = lambda option: option  # type: ignore
    quirks.read(quirksfile)

    for name, section in filter(lambda n: n != "DEFAULT", quirks.items()):
        vid = int(section.get("MatchVendor", "0x0"), 16)
        if vid == 0x046D:
            pid = int(section.get("MatchProduct", "0x0"), 16)
            assert pid not in logitech_receivers, (
                f"{quirksfile}: {name}: applies to a Logitech Receiver"
            )


def main():
    args = [__file__]
    try:
        import xdist  # noqa

        ncores = os.environ.get("FDO_CI_CONCURRENT", "auto")
        args += ["-n", ncores]
    except ImportError:
        pass

    return pytest.main(args)


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
# -*- coding: utf-8
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2021 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the 'Software'),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#
# Prints the data from a libinput recording in a table format to ease
# debugging.
#
# Input is a libinput record yaml file

import argparse
import os
import sys
import yaml
import libevdev

# minimum width of a field in the table
MIN_FIELD_WIDTH = 6


# Default is to just return the value of an axis, but some axes want special
# formatting.
def format_value(code, value):
    if code in (libevdev.EV_ABS.ABS_MISC, libevdev.EV_MSC.MSC_SERIAL):
        return f"{value & 0xFFFFFFFF:#x}"

    # Rel axes we always print the sign
    if code.type == libevdev.EV_REL:
        return f"{value:+d}"

    return f"{value}"


# The list of axes we want to track
def is_tracked_axis(code, allowlist, denylist):
    if code.type in (libevdev.EV_KEY, libevdev.EV_SW, libevdev.EV_SYN):
        return False

    # We don't do slots in this tool
    if code.type == libevdev.EV_ABS:
        if libevdev.EV_ABS.ABS_MT_SLOT <= code <= libevdev.EV_ABS.ABS_MAX:
            return False

    if allowlist:
        return code in allowlist
    else:
        return code not in denylist


def main(argv):
    parser = argparse.ArgumentParser(
        description="Display a recording in a tabular format"
    )
    parser.add_argument(
        "path", metavar="recording", nargs=1, help="Path to libinput-record YAML file"
    )
    parser.add_argument(
        "--ignore",
        metavar="ABS_X,ABS_Y,...",
        default="",
        help="A comma-separated list of axis names to ignore",
    )
    parser.add_argument(
        "--only",
        metavar="ABS_X,ABS_Y,...",
        default="",
        help="A comma-separated list of axis names to print, ignoring all others",
    )
    parser.add_argument(
        "--print-state",
        action="store_true",
        default=False,
        help="Always print all axis values, even unchanged ones",
    )

    args = parser.parse_args()
    if args.ignore and args.only:
        print("Only one of --ignore and --only may be given", file=sys.stderr)
        sys.exit(2)

    ignored_axes = [libevdev.evbit(axis) for axis in args.ignore.split(",") if axis]
    only_axes = [libevdev.evbit(axis) for axis in args.only.split(",") if axis]

    isatty = os.isatty(sys.stdout.fileno())

    yml = yaml.safe_load(open(args.path[0]))
    if yml["ndevices"] > 1:
        print(f"WARNING: Using only first {yml['ndevices']} devices in recording")
    device = yml["devices"][0]

    if not device["events"]:
        print("No events found in recording")
        sys.exit(1)

    def events():
        """
        Yields the next event in the recording
        """
        for event in device["events"]:
            for evdev in event.get("evdev", []):
                yield libevdev.InputEvent(
                    code=libevdev.evbit(evdev[2], evdev[3]),
                    value=evdev[4],
                    sec=evdev[0],
                    usec=evdev[1],
                )

    def interesting_axes(events):
        """
        Yields the libevdev codes with the axes in this recording
        """
        used_axes = []
        for e in events:
            if e.code not in used_axes and is_tracked_axis(
                e.code, only_axes, ignored_axes
            ):
                yield e.code
                used_axes.append(e.code)

    # Compile all axes that we want to print first
    axes = sorted(
        interesting_axes(events()), key=lambda x: x.type.value * 1000 + x.value
    )
    # Strip the REL_/ABS_ prefix for the headers
    headers = [a.name[4:].rjust(MIN_FIELD_WIDTH) for a in axes]
    # for easier formatting later, we keep the header field width in a dict
    axes = {a: len(h) for a, h in zip(axes, headers)}

    # Time is a special case, always the first entry
    # Format uses ms only, we rarely ever care about µs
    headers = [f"{'Time':<7s}"] + headers + ["Keys"]
    header_line = f"{' | '.join(headers)}"
    print(header_line)
    print("-" * len(header_line))

    current_codes = []
    current_frame = {}  # {evdev-code: value}
    axes_in_use = {}  # to print axes never sending events
    last_fields = []  # to skip duplicate lines
    continuation_count = 0

    keystate = {}
    keystate_changed = False

    for e in events():
        axes_in_use[e.code] = True

        if e.code.type == libevdev.EV_KEY:
            keystate[e.code] = e.value
            keystate_changed = True
        elif is_tracked_axis(e.code, only_axes, ignored_axes):
            current_frame[e.code] = e.value
            current_codes.append(e.code)
        elif e.code == libevdev.EV_SYN.SYN_REPORT:
            fields = []
            for a in axes:
                if args.print_state or a in current_codes:
                    s = format_value(a, current_frame.get(a, 0))
                else:
                    s = ""
                fields.append(s.rjust(max(MIN_FIELD_WIDTH, axes[a])))
            current_codes = []

            if last_fields != fields or keystate_changed:
                last_fields = fields.copy()
                keystate_changed = False

                if continuation_count:
                    if not isatty:
                        print(f" ... +{continuation_count}", end="")
                    print("")
                    continuation_count = 0

                fields.insert(0, f"{e.sec: 3d}.{e.usec // 1000:03d}")
                keys_down = [k.name for k, v in keystate.items() if v]
                fields.append(", ".join(keys_down))
                print(" | ".join(fields))
            else:
                continuation_count += 1
                if isatty:
                    print(f"\r ... +{continuation_count}", end="", flush=True)

    # Print out any rel/abs axes that not generate events in
    # this recording
    unused_axes = []
    for evtype, evcodes in device["evdev"]["codes"].items():
        for c in evcodes:
            code = libevdev.evbit(int(evtype), int(c))
            if (
                is_tracked_axis(code, only_axes, ignored_axes)
                and code not in axes_in_use
            ):
                unused_axes.append(code)

    if unused_axes:
        print(
            f"Axes present but without events: {', '.join([a.name for a in unused_axes])}"
        )

    for e in events():
        if libevdev.EV_ABS.ABS_MT_SLOT <= code <= libevdev.EV_ABS.ABS_MAX:
            print(
                "WARNING: This recording contains multitouch data that is not supported by this tool."
            )
            break


if __name__ == "__main__":
    try:
        main(sys.argv)
    except BrokenPipeError:
        pass



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import argparse
import os
import sys
import unittest
import yaml
import re

from pkg_resources import parse_version


class TestYaml(unittest.TestCase):
    filename = ""

    @classmethod
    def setUpClass(cls):
        with open(cls.filename) as f:
            cls.yaml = yaml.safe_load(f)

    def dict_key_crosscheck(self, d, keys):
        """Check that each key in d is in keys, and that each key is in d"""
        self.assertEqual(sorted(d.keys()), sorted(keys))

    def libinput_events(self, filter=None):
        """Returns all libinput events in the recording, regardless of the
        device"""
        devices = self.yaml["devices"]
        for d in devices:
            events = d["events"]
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    libinput = e["libinput"]
                except KeyError:
                    continue

                for ev in libinput:
                    if (
                        filter is None
                        or ev["type"] == filter
                        or isinstance(filter, list)
                        and ev["type"] in filter
                    ):
                        yield ev

    def test_sections_exist(self):
        sections = ["version", "ndevices", "libinput", "system", "devices"]
        for section in sections:
            self.assertIn(section, self.yaml)

    def test_version(self):
        version = self.yaml["version"]
        self.assertTrue(isinstance(version, int))
        self.assertEqual(version, 1)

    def test_ndevices(self):
        ndevices = self.yaml["ndevices"]
        self.assertTrue(isinstance(ndevices, int))
        self.assertGreaterEqual(ndevices, 1)
        self.assertEqual(ndevices, len(self.yaml["devices"]))

    def test_libinput(self):
        libinput = self.yaml["libinput"]
        version = libinput["version"]
        self.assertTrue(isinstance(version, str))
        self.assertGreaterEqual(parse_version(version), parse_version("1.10.0"))
        git = libinput["git"]
        self.assertTrue(isinstance(git, str))
        self.assertNotEqual(git, "unknown")

    def test_system(self):
        system = self.yaml["system"]
        kernel = system["kernel"]
        self.assertTrue(isinstance(kernel, str))
        self.assertEqual(kernel, os.uname().release)

        dmi = system["dmi"]
        self.assertTrue(isinstance(dmi, str))
        with open("/sys/class/dmi/id/modalias") as f:
            sys_dmi = f.read()[:-1]  # trailing newline
            self.assertEqual(dmi, sys_dmi)

    def test_devices_sections_exist(self):
        devices = self.yaml["devices"]
        for d in devices:
            self.assertIn("node", d)
            self.assertIn("evdev", d)
            self.assertIn("udev", d)

    def test_evdev_sections_exist(self):
        sections = ["name", "id", "codes", "properties"]
        devices = self.yaml["devices"]
        for d in devices:
            evdev = d["evdev"]
            for s in sections:
                self.assertIn(s, evdev)

    def test_evdev_name(self):
        devices = self.yaml["devices"]
        for d in devices:
            evdev = d["evdev"]
            name = evdev["name"]
            self.assertTrue(isinstance(name, str))
            self.assertGreaterEqual(len(name), 5)

    def test_evdev_id(self):
        devices = self.yaml["devices"]
        for d in devices:
            evdev = d["evdev"]
            id = evdev["id"]
            self.assertTrue(isinstance(id, list))
            self.assertEqual(len(id), 4)
            self.assertGreater(id[0], 0)
            self.assertGreater(id[1], 0)

    def test_evdev_properties(self):
        devices = self.yaml["devices"]
        for d in devices:
            evdev = d["evdev"]
            properties = evdev["properties"]
            self.assertTrue(isinstance(properties, list))

    def test_hid(self):
        devices = self.yaml["devices"]
        for d in devices:
            hid = d["hid"]
            self.assertTrue(isinstance(hid, list))
            for byte in hid:
                self.assertGreaterEqual(byte, 0)
                self.assertLessEqual(byte, 255)

    def test_udev_sections_exist(self):
        sections = ["properties"]
        devices = self.yaml["devices"]
        for d in devices:
            udev = d["udev"]
            for s in sections:
                self.assertIn(s, udev)

    def test_udev_properties(self):
        devices = self.yaml["devices"]
        for d in devices:
            udev = d["udev"]
            properties = udev["properties"]
            self.assertTrue(isinstance(properties, list))
            self.assertGreater(len(properties), 0)

            self.assertIn("ID_INPUT=1", properties)
            for p in properties:
                self.assertTrue(re.match("[A-Z0-9_]+=.+", p))

    def test_udev_id_inputs(self):
        devices = self.yaml["devices"]
        for d in devices:
            udev = d["udev"]
            properties = udev["properties"]
            id_inputs = [p for p in properties if p.startswith("ID_INPUT")]
            # We expect ID_INPUT and ID_INPUT_something, but might get more
            # than one of the latter
            self.assertGreaterEqual(len(id_inputs), 2)

    def test_events_have_section(self):
        devices = self.yaml["devices"]
        for d in devices:
            events = d["events"]
            if not events:
                raise unittest.SkipTest()
            for e in events:
                self.assertTrue("evdev" in e or "libinput" in e)

    def test_events_evdev(self):
        devices = self.yaml["devices"]
        for d in devices:
            events = d["events"]
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    evdev = e["evdev"]
                except KeyError:
                    continue

                for ev in evdev:
                    self.assertEqual(len(ev), 5)

                # Last event in each frame is SYN_REPORT
                ev_syn = evdev[-1]
                self.assertEqual(ev_syn[2], 0)
                self.assertEqual(ev_syn[3], 0)
                # SYN_REPORT value is 1 in case of some key repeats
                self.assertLessEqual(ev_syn[4], 1)

    def test_events_evdev_syn_report(self):
        devices = self.yaml["devices"]
        for d in devices:
            events = d["events"]
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    evdev = e["evdev"]
                except KeyError:
                    continue
                for ev in evdev[:-1]:
                    self.assertFalse(ev[2] == 0 and ev[3] == 0)

    def test_events_libinput(self):
        devices = self.yaml["devices"]
        for d in devices:
            events = d["events"]
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    libinput = e["libinput"]
                except KeyError:
                    continue

                self.assertTrue(isinstance(libinput, list))
                for ev in libinput:
                    self.assertTrue(isinstance(ev, dict))

    def test_events_libinput_type(self):
        types = [
            "POINTER_MOTION",
            "POINTER_MOTION_ABSOLUTE",
            "POINTER_AXIS",
            "POINTER_BUTTON",
            "DEVICE_ADDED",
            "KEYBOARD_KEY",
            "TOUCH_DOWN",
            "TOUCH_MOTION",
            "TOUCH_UP",
            "TOUCH_FRAME",
            "GESTURE_SWIPE_BEGIN",
            "GESTURE_SWIPE_UPDATE",
            "GESTURE_SWIPE_END",
            "GESTURE_PINCH_BEGIN",
            "GESTURE_PINCH_UPDATE",
            "GESTURE_PINCH_END",
            "TABLET_TOOL_AXIS",
            "TABLET_TOOL_PROXIMITY",
            "TABLET_TOOL_BUTTON",
            "TABLET_TOOL_TIP",
            "TABLET_PAD_STRIP",
            "TABLET_PAD_RING",
            "TABLET_PAD_BUTTON",
            "SWITCH_TOGGLE",
        ]
        for e in self.libinput_events():
            self.assertIn("type", e)
            self.assertIn(e["type"], types)

    def test_events_libinput_time(self):
        # DEVICE_ADDED has no time
        # first event may have 0.0 time if the first frame generates a
        # libinput event.
        try:
            for e in list(self.libinput_events())[2:]:
                self.assertIn("time", e)
                self.assertGreater(e["time"], 0.0)
                self.assertLess(e["time"], 60.0)
        except IndexError:
            pass

    def test_events_libinput_device_added(self):
        keys = ["type", "seat", "logical_seat"]
        for e in self.libinput_events("DEVICE_ADDED"):
            self.dict_key_crosscheck(e, keys)
            self.assertEqual(e["seat"], "seat0")
            self.assertEqual(e["logical_seat"], "default")

    def test_events_libinput_pointer_motion(self):
        keys = ["type", "time", "delta", "unaccel"]
        for e in self.libinput_events("POINTER_MOTION"):
            self.dict_key_crosscheck(e, keys)
            delta = e["delta"]
            self.assertTrue(isinstance(delta, list))
            self.assertEqual(len(delta), 2)
            for d in delta:
                self.assertTrue(isinstance(d, float))
            unaccel = e["unaccel"]
            self.assertTrue(isinstance(unaccel, list))
            self.assertEqual(len(unaccel), 2)
            for d in unaccel:
                self.assertTrue(isinstance(d, float))

    def test_events_libinput_pointer_button(self):
        keys = ["type", "time", "button", "state", "seat_count"]
        for e in self.libinput_events("POINTER_BUTTON"):
            self.dict_key_crosscheck(e, keys)
            button = e["button"]
            self.assertGreater(button, 0x100)  # BTN_0
            self.assertLess(button, 0x160)  # KEY_OK
            state = e["state"]
            self.assertIn(state, ["pressed", "released"])
            scount = e["seat_count"]
            self.assertGreaterEqual(scount, 0)

    def test_events_libinput_pointer_absolute(self):
        keys = ["type", "time", "point", "transformed"]
        for e in self.libinput_events("POINTER_MOTION_ABSOLUTE"):
            self.dict_key_crosscheck(e, keys)
            point = e["point"]
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)
                self.assertLess(p, 300.0)

            transformed = e["transformed"]
            self.assertTrue(isinstance(transformed, list))
            self.assertEqual(len(transformed), 2)
            for t in transformed:
                self.assertTrue(isinstance(t, float))
                self.assertGreater(t, 0.0)
                self.assertLess(t, 100.0)

    def test_events_libinput_touch(self):
        keys = ["type", "time", "slot", "seat_slot"]
        for e in self.libinput_events():
            if not e["type"].startswith("TOUCH_") or e["type"] == "TOUCH_FRAME":
                continue

            for k in keys:
                self.assertIn(k, e.keys())
            slot = e["slot"]
            seat_slot = e["seat_slot"]

            self.assertGreaterEqual(slot, 0)
            self.assertGreaterEqual(seat_slot, 0)

    def test_events_libinput_touch_down(self):
        keys = ["type", "time", "slot", "seat_slot", "point", "transformed"]
        for e in self.libinput_events("TOUCH_DOWN"):
            self.dict_key_crosscheck(e, keys)
            point = e["point"]
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)
                self.assertLess(p, 300.0)

            transformed = e["transformed"]
            self.assertTrue(isinstance(transformed, list))
            self.assertEqual(len(transformed), 2)
            for t in transformed:
                self.assertTrue(isinstance(t, float))
                self.assertGreater(t, 0.0)
                self.assertLess(t, 100.0)

    def test_events_libinput_touch_motion(self):
        keys = ["type", "time", "slot", "seat_slot", "point", "transformed"]
        for e in self.libinput_events("TOUCH_MOTION"):
            self.dict_key_crosscheck(e, keys)
            point = e["point"]
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)
                self.assertLess(p, 300.0)

            transformed = e["transformed"]
            self.assertTrue(isinstance(transformed, list))
            self.assertEqual(len(transformed), 2)
            for t in transformed:
                self.assertTrue(isinstance(t, float))
                self.assertGreater(t, 0.0)
                self.assertLess(t, 100.0)

    def test_events_libinput_touch_frame(self):
        devices = self.yaml["devices"]
        for d in devices:
            events = d["events"]
            if not events:
                raise unittest.SkipTest()
            for e in events:
                try:
                    evdev = e["libinput"]
                except KeyError:
                    continue

                need_frame = False
                for ev in evdev:
                    t = ev["type"]
                    if not t.startswith("TOUCH_"):
                        self.assertFalse(need_frame)
                        continue

                    if t == "TOUCH_FRAME":
                        self.assertTrue(need_frame)
                        need_frame = False
                    else:
                        need_frame = True

                self.assertFalse(need_frame)

    def test_events_libinput_gesture_pinch(self):
        keys = ["type", "time", "nfingers", "delta", "unaccel", "angle_delta", "scale"]
        for e in self.libinput_events(
            ["GESTURE_PINCH_BEGIN", "GESTURE_PINCH_UPDATE", "GESTURE_PINCH_END"]
        ):
            self.dict_key_crosscheck(e, keys)
            delta = e["delta"]
            self.assertTrue(isinstance(delta, list))
            self.assertEqual(len(delta), 2)
            for d in delta:
                self.assertTrue(isinstance(d, float))
            unaccel = e["unaccel"]
            self.assertTrue(isinstance(unaccel, list))
            self.assertEqual(len(unaccel), 2)
            for d in unaccel:
                self.assertTrue(isinstance(d, float))

            adelta = e["angle_delta"]
            self.assertTrue(isinstance(adelta, list))
            self.assertEqual(len(adelta), 2)
            for d in adelta:
                self.assertTrue(isinstance(d, float))

            scale = e["scale"]
            self.assertTrue(isinstance(scale, list))
            self.assertEqual(len(scale), 2)
            for d in scale:
                self.assertTrue(isinstance(d, float))

    def test_events_libinput_gesture_swipe(self):
        keys = ["type", "time", "nfingers", "delta", "unaccel"]
        for e in self.libinput_events(
            ["GESTURE_SWIPE_BEGIN", "GESTURE_SWIPE_UPDATE", "GESTURE_SWIPE_END"]
        ):
            self.dict_key_crosscheck(e, keys)
            delta = e["delta"]
            self.assertTrue(isinstance(delta, list))
            self.assertEqual(len(delta), 2)
            for d in delta:
                self.assertTrue(isinstance(d, float))
            unaccel = e["unaccel"]
            self.assertTrue(isinstance(unaccel, list))
            self.assertEqual(len(unaccel), 2)
            for d in unaccel:
                self.assertTrue(isinstance(d, float))

    def test_events_libinput_tablet_pad_button(self):
        keys = ["type", "time", "button", "state", "mode", "is-toggle"]

        for e in self.libinput_events("TABLET_PAD_BUTTON"):
            self.dict_key_crosscheck(e, keys)

            b = e["button"]
            self.assertTrue(isinstance(b, int))
            self.assertGreaterEqual(b, 0)
            self.assertLessEqual(b, 16)

            state = e["state"]
            self.assertIn(state, ["pressed", "released"])

            m = e["mode"]
            self.assertTrue(isinstance(m, int))
            self.assertGreaterEqual(m, 0)
            self.assertLessEqual(m, 3)

            t = e["is-toggle"]
            self.assertTrue(isinstance(t, bool))

    def test_events_libinput_tablet_pad_ring(self):
        keys = ["type", "time", "number", "position", "source", "mode"]

        for e in self.libinput_events("TABLET_PAD_RING"):
            self.dict_key_crosscheck(e, keys)

            n = e["number"]
            self.assertTrue(isinstance(n, int))
            self.assertGreaterEqual(n, 0)
            self.assertLessEqual(n, 4)

            p = e["position"]
            self.assertTrue(isinstance(p, float))
            if p != -1.0:  # special 'end' case
                self.assertGreaterEqual(p, 0.0)
                self.assertLess(p, 360.0)

            m = e["mode"]
            self.assertTrue(isinstance(m, int))
            self.assertGreaterEqual(m, 0)
            self.assertLessEqual(m, 3)

            s = e["source"]
            self.assertIn(s, ["finger", "unknown"])

    def test_events_libinput_tablet_pad_strip(self):
        keys = ["type", "time", "number", "position", "source", "mode"]

        for e in self.libinput_events("TABLET_PAD_STRIP"):
            self.dict_key_crosscheck(e, keys)

            n = e["number"]
            self.assertTrue(isinstance(n, int))
            self.assertGreaterEqual(n, 0)
            self.assertLessEqual(n, 4)

            p = e["position"]
            self.assertTrue(isinstance(p, float))
            if p != -1.0:  # special 'end' case
                self.assertGreaterEqual(p, 0.0)
                self.assertLessEqual(p, 1.0)

            m = e["mode"]
            self.assertTrue(isinstance(m, int))
            self.assertGreaterEqual(m, 0)
            self.assertLessEqual(m, 3)

            s = e["source"]
            self.assertIn(s, ["finger", "unknown"])

    def test_events_libinput_tablet_tool_proximity(self):
        keys = ["type", "time", "proximity", "tool-type", "serial", "axes"]

        for e in self.libinput_events("TABLET_TOOL_PROXIMITY"):
            for k in keys:
                self.assertIn(k, e)

            p = e["proximity"]
            self.assertIn(p, ["in", "out"])

            p = e["tool-type"]
            self.assertIn(
                p, ["pen", "eraser", "brush", "airbrush", "mouse", "lens", "unknown"]
            )

            s = e["serial"]
            self.assertTrue(isinstance(s, int))
            self.assertGreaterEqual(s, 0)

            a = e["axes"]
            for ax in e["axes"]:
                self.assertIn(a, "pdtrsw")

    def test_events_libinput_tablet_tool(self):
        keys = ["type", "time", "tip"]

        for e in self.libinput_events(["TABLET_TOOL_AXIS", "TABLET_TOOL_TIP"]):
            for k in keys:
                self.assertIn(k, e)

            t = e["tip"]
            self.assertIn(t, ["down", "up"])

    def test_events_libinput_tablet_tool_button(self):
        keys = ["type", "time", "button", "state"]

        for e in self.libinput_events("TABLET_TOOL_BUTTON"):
            self.dict_key_crosscheck(e, keys)

            b = e["button"]
            # STYLUS, STYLUS2, STYLUS3
            self.assertIn(b, [0x14B, 0x14C, 0x139])

            s = e["state"]
            self.assertIn(s, ["pressed", "released"])

    def test_events_libinput_tablet_tool_axes(self):
        for e in self.libinput_events(
            ["TABLET_TOOL_PROXIMITY", "TABLET_TOOL_AXIS", "TABLET_TOOL_TIP"]
        ):
            point = e["point"]
            self.assertTrue(isinstance(point, list))
            self.assertEqual(len(point), 2)
            for p in point:
                self.assertTrue(isinstance(p, float))
                self.assertGreater(p, 0.0)

            try:
                tilt = e["tilt"]
                self.assertTrue(isinstance(tilt, list))
                self.assertEqual(len(tilt), 2)
                for t in tilt:
                    self.assertTrue(isinstance(t, float))
            except KeyError:
                pass

            try:
                d = e["distance"]
                self.assertTrue(isinstance(d, float))
                self.assertGreaterEqual(d, 0.0)
                self.assertNotIn("pressure", e)
            except KeyError:
                pass

            try:
                p = e["pressure"]
                self.assertTrue(isinstance(p, float))
                self.assertGreaterEqual(p, 0.0)
                self.assertNotIn("distance", e)
            except KeyError:
                pass

            try:
                r = e["rotation"]
                self.assertTrue(isinstance(r, float))
                self.assertGreaterEqual(r, 0.0)
            except KeyError:
                pass

            try:
                s = e["slider"]
                self.assertTrue(isinstance(s, float))
                self.assertGreaterEqual(s, 0.0)
            except KeyError:
                pass

            try:
                w = e["wheel"]
                self.assertTrue(isinstance(w, float))
                self.assertGreaterEqual(w, 0.0)
                self.assertIn("wheel-discrete", e)
                wd = e["wheel-discrete"]
                self.assertTrue(isinstance(wd, 1))
                self.assertGreaterEqual(wd, 0.0)

                def sign(x):
                    (1, -1)[x < 0]

                self.assertTrue(sign(w), sign(wd))
            except KeyError:
                pass

    def test_events_libinput_switch(self):
        keys = ["type", "time", "switch", "state"]

        for e in self.libinput_events("SWITCH_TOGGLE"):
            self.dict_key_crosscheck(e, keys)

            s = e["switch"]
            self.assertTrue(isinstance(s, int))
            self.assertIn(s, [0x00, 0x01])

            # yaml converts on/off to true/false
            state = e["state"]
            self.assertTrue(isinstance(state, bool))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Verify a YAML recording")
    parser.add_argument(
        "recording",
        metavar="recorded-file.yaml",
        type=str,
        help="Path to device recording",
    )
    parser.add_argument("--verbose", action="store_true")
    args, remainder = parser.parse_known_args()
    TestYaml.filename = args.recording
    verbosity = 1
    if args.verbose:
        verbosity = 3

    argv = [sys.argv[0], *remainder]
    unittest.main(argv=argv, verbosity=verbosity)



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
import resource
import sys
import subprocess
import logging

try:
    import pytest
except ImportError:
    print("Failed to import pytest. Skipping.", file=sys.stderr)
    sys.exit(77)


logger = logging.getLogger("test")
logger.setLevel(logging.DEBUG)

if "@DISABLE_WARNING@" != "yes":
    print("This is the source file, run the one in the meson builddir instead")
    sys.exit(1)


def _disable_coredump():
    resource.setrlimit(resource.RLIMIT_CORE, (0, 0))


def run_command(args):
    logger.debug("run command: {}".format(" ".join(args)))
    with subprocess.Popen(
        args,
        preexec_fn=_disable_coredump,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as p:
        try:
            p.wait(0.7)
        except subprocess.TimeoutExpired:
            p.send_signal(3)  # SIGQUIT
        stdout, stderr = p.communicate(timeout=5)
        if p.returncode == -3:
            p.returncode = 0
        return p.returncode, stdout.decode("UTF-8"), stderr.decode("UTF-8")


class LibinputTool(object):
    libinput_tool = "libinput"
    subtool = None

    def __init__(self, subtool=None):
        self.libinput_tool = "@TOOL_PATH@"
        self.subtool = subtool

    def run_command(self, args):
        args = [self.libinput_tool] + args
        if self.subtool is not None:
            args.insert(1, self.subtool)

        return run_command(args)

    def run_command_success(self, args):
        rc, stdout, stderr = self.run_command(args)
        # if we're running as user, we might fail the command but we should
        # never get rc 2 (invalid usage)
        assert rc in [0, 1], (stdout, stderr)
        return stdout, stderr

    def run_command_invalid(self, args):
        rc, stdout, stderr = self.run_command(args)
        assert rc == 2, (rc, stdout, stderr)
        return rc, stdout, stderr

    def run_command_unrecognized_option(self, args):
        rc, stdout, stderr = self.run_command(args)
        assert rc == 2, (rc, stdout, stderr)
        assert stdout.startswith("Usage") or stdout == ""
        assert "unrecognized option" in stderr

    def run_command_missing_arg(self, args):
        rc, stdout, stderr = self.run_command(args)
        assert rc == 2, (rc, stdout, stderr)
        assert stdout.startswith("Usage") or stdout == ""
        assert "requires an argument" in stderr

    def run_command_unrecognized_tool(self, args):
        rc, stdout, stderr = self.run_command(args)
        assert rc == 2, (rc, stdout, stderr)
        assert stdout.startswith("Usage") or stdout == ""
        assert "is not installed" in stderr


class LibinputDebugGui(LibinputTool):
    def __init__(self, subtool="debug-gui"):
        assert subtool == "debug-gui"
        super().__init__(subtool)

        debug_gui_enabled = "@MESON_ENABLED_DEBUG_GUI@" == "True"
        if not debug_gui_enabled:
            pytest.skip()

        if not os.getenv("DISPLAY") and not os.getenv("WAYLAND_DISPLAY"):
            pytest.skip()

        # 77 means gtk_init() failed, which is probably because you can't
        # connect to the display server.
        rc, _, _ = self.run_command(["--help"])
        if rc == 77:
            pytest.skip()


def get_tool(subtool=None):
    if subtool == "debug-gui":
        return LibinputDebugGui()
    else:
        return LibinputTool(subtool)


@pytest.fixture
def libinput():
    return get_tool()


@pytest.fixture(params=["debug-events", "debug-gui"])
def libinput_debug_tool(request):
    yield get_tool(request.param)


@pytest.fixture
def libinput_debug_events():
    return get_tool("debug-events")


@pytest.fixture
def libinput_debug_gui():
    return get_tool("debug-gui")


@pytest.fixture
def libinput_record():
    return get_tool("record")


def test_help(libinput):
    stdout, stderr = libinput.run_command_success(["--help"])
    assert stdout.startswith("Usage:")
    assert stderr == ""


def test_version(libinput):
    stdout, stderr = libinput.run_command_success(["--version"])
    assert stdout.startswith("1")
    assert stderr == ""


@pytest.mark.parametrize("argument", ["--banana", "--foo", "--quiet", "--verbose"])
def test_invalid_arguments(libinput, argument):
    libinput.run_command_unrecognized_option([argument])


@pytest.mark.parametrize("tool", [["foo"], ["debug"], ["foo", "--quiet"]])
def test_invalid_tool(libinput, tool):
    libinput.run_command_unrecognized_tool(tool)


def test_udev_seat(libinput_debug_tool):
    libinput_debug_tool.run_command_missing_arg(["--udev"])
    libinput_debug_tool.run_command_success(["--udev", "seat0"])
    libinput_debug_tool.run_command_success(["--udev", "seat1"])


@pytest.mark.skipif(os.environ.get("UDEV_NOT_AVAILABLE"), reason="udev required")
def test_device_arg(libinput_debug_tool):
    libinput_debug_tool.run_command_missing_arg(["--device"])
    libinput_debug_tool.run_command_success(["--device", "/dev/input/event0"])
    libinput_debug_tool.run_command_success(["--device", "/dev/input/event1"])
    libinput_debug_tool.run_command_success(["/dev/input/event0"])


options = {
    "pattern": ["sendevents"],
    # enable/disable options
    "enable-disable": [
        "tap",
        "drag",
        "drag-lock",
        "middlebutton",
        "natural-scrolling",
        "left-handed",
        "dwt",
        "dwtp",
    ],
    # options with distinct values
    "enums": {
        "set-click-method": ["none", "clickfinger", "buttonareas"],
        "set-scroll-method": ["none", "twofinger", "edge", "button"],
        "set-profile": ["adaptive", "flat"],
        "set-tap-map": ["lrm", "lmr"],
        "set-clickfinger-map": ["lrm", "lmr"],
        "enable-drag-lock": ["sticky", "timeout"],
    },
    # options with a range (and increment)
    "ranges": {
        "set-speed": (-1.0, +1.0, 0.1),
        "set-rotation": (0, 360, 10),
    },
}


# Options that allow for glob patterns
@pytest.mark.parametrize("option", options["pattern"])
def test_options_pattern(libinput_debug_tool, option):
    libinput_debug_tool.run_command_success(["--disable-{}".format(option), "*"])
    libinput_debug_tool.run_command_success(["--disable-{}".format(option), "abc*"])


@pytest.mark.parametrize("option", options["enable-disable"])
def test_options_enable_disable(libinput_debug_tool, option):
    libinput_debug_tool.run_command_success(["--enable-{}".format(option)])
    libinput_debug_tool.run_command_success(["--disable-{}".format(option)])


@pytest.mark.parametrize("option", options["enums"].items())
def test_options_enums(libinput_debug_tool, option):
    name, values = option
    for v in values:
        libinput_debug_tool.run_command_success(["--{}".format(name), v])
        libinput_debug_tool.run_command_success(["--{}={}".format(name, v)])


@pytest.mark.parametrize("option", options["ranges"].items())
def test_options_ranges(libinput_debug_tool, option):
    name, values = option
    minimum, maximum, step = values
    value = minimum
    while value < maximum:
        libinput_debug_tool.run_command_success(["--{}".format(name), str(value)])
        libinput_debug_tool.run_command_success(["--{}={}".format(name, value)])
        value += step
    libinput_debug_tool.run_command_success(["--{}".format(name), str(maximum)])
    libinput_debug_tool.run_command_success(["--{}={}".format(name, maximum)])


def test_apply_to(libinput_debug_tool):
    libinput_debug_tool.run_command_missing_arg(["--apply-to"])
    libinput_debug_tool.run_command_success(["--apply-to", "*foo*"])
    libinput_debug_tool.run_command_success(["--apply-to", "foobar"])
    libinput_debug_tool.run_command_success(["--apply-to", "any"])


@pytest.mark.parametrize(
    "args",
    [["--verbose"], ["--quiet"], ["--verbose", "--quiet"], ["--quiet", "--verbose"]],
)
def test_debug_events_verbose_quiet(libinput_debug_events, args):
    libinput_debug_events.run_command_success(args)


@pytest.mark.parametrize("arg", ["--banana", "--foo", "--version"])
def test_invalid_args(libinput_debug_tool, arg):
    libinput_debug_tool.run_command_unrecognized_option([arg])


def test_libinput_debug_events_multiple_devices(libinput_debug_events):
    libinput_debug_events.run_command_success(
        ["--device", "/dev/input/event0", "/dev/input/event1"]
    )
    # same event path multiple times? meh, your problem
    libinput_debug_events.run_command_success(
        ["--device", "/dev/input/event0", "/dev/input/event0"]
    )
    libinput_debug_events.run_command_success(
        ["/dev/input/event0", "/dev/input/event1"]
    )


def test_libinput_debug_events_too_many_devices(libinput_debug_events):
    # Too many arguments just bails with the usage message
    rc, stdout, stderr = libinput_debug_events.run_command(["/dev/input/event0"] * 61)
    assert rc == 2, (stdout, stderr)


@pytest.mark.parametrize("arg", ["--quiet"])
def test_libinput_debug_gui_invalid_arg(libinput_debug_gui, arg):
    libinput_debug_gui.run_command_unrecognized_option([arg])


def test_libinput_debug_gui_verbose(libinput_debug_gui):
    libinput_debug_gui.run_command_success(["--verbose"])


@pytest.mark.parametrize(
    "arg", ["--help", "--show-keycodes", "--with-libinput", "--with-hidraw"]
)
def test_libinput_record_args(libinput_record, arg):
    libinput_record.run_command_success([arg])


def test_libinput_record_multiple_arg(libinput_record):
    # this arg is deprecated and a noop
    libinput_record.run_command_success(["--multiple"])


@pytest.fixture
def recording(tmp_path):
    return str((tmp_path / "record.out").resolve())


def test_libinput_record_all(libinput_record, recording):
    libinput_record.run_command_success(["--all", "-o", recording])
    libinput_record.run_command_success(["--all", recording])


def test_libinput_record_outfile(libinput_record, recording):
    libinput_record.run_command_success(["-o", recording])
    libinput_record.run_command_success(["--output-file", recording])
    libinput_record.run_command_success(["--output-file={}".format(recording)])


def test_libinput_record_single(libinput_record, recording):
    libinput_record.run_command_success(["/dev/input/event0"])
    libinput_record.run_command_success(["-o", recording, "/dev/input/event0"])
    libinput_record.run_command_success(["/dev/input/event0", recording])
    libinput_record.run_command_success([recording, "/dev/input/event0"])


def test_libinput_record_multiple(libinput_record, recording):
    libinput_record.run_command_success(
        ["-o", recording, "/dev/input/event0", "/dev/input/event1"]
    )
    libinput_record.run_command_success(
        [recording, "/dev/input/event0", "/dev/input/event1"]
    )
    libinput_record.run_command_success(
        ["/dev/input/event0", "/dev/input/event1", recording]
    )


def test_libinput_record_autorestart(libinput_record, recording):
    libinput_record.run_command_invalid(["--autorestart"])
    libinput_record.run_command_invalid(["--autorestart=2"])
    libinput_record.run_command_success(["-o", recording, "--autorestart=2"])


def main():
    args = ["-m", "pytest"]
    try:
        import xdist  # noqa

        ncores = os.environ.get("FDO_CI_CONCURRENT", "auto")
        args += ["-n", ncores]
    except ImportError:
        logger.info("python-xdist missing, this test will be slow")
        pass

    args += ["@MESON_BUILD_ROOT@"]

    os.environ["LIBINPUT_RUNNING_TEST_SUITE"] = "1"

    return subprocess.run([sys.executable] + args).returncode


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2017 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#

import sys
import subprocess
import argparse
import fcntl
import os
import select
import termios
import tty

try:
    import libevdev
    import pyudev
except ModuleNotFoundError as e:
    print("Error: {}".format(str(e)), file=sys.stderr)
    print(
        "One or more python modules are missing. Please install those "
        "modules and re-run this tool."
    )
    sys.exit(1)


class TableFormatter(object):
    ALIGNMENT = 3

    def __init__(self):
        self.colwidths = []

    @property
    def width(self):
        return sum(self.colwidths) + 1

    def headers(self, args):
        s = "│"
        align = self.ALIGNMENT - 1  # account for │

        for arg in args:
            # +2 because we want space left/right of text
            w = ((len(arg) + 2 + align) // align) * align
            self.colwidths.append(w + 1)
            s += " {:^{width}s} │".format(arg, width=w - 2)

        return s

    def values(self, args):
        s = "│"
        for w, arg in zip(self.colwidths, args):
            w -= 1  # width includes │ separator
            if isinstance(arg, str):
                # We want space margins for strings
                s += " {:{width}s} │".format(arg, width=w - 2)
            elif isinstance(arg, bool):
                s += "{:^{width}s}│".format("x" if arg else " ", width=w)
            else:
                s += "{:^{width}d}│".format(arg, width=w)

        if len(args) < len(self.colwidths):
            s += "│".rjust(self.width - len(s), " ")
        return s

    def header(self):
        return "┌" + "─" * (self.width - 2) + "┐"

    def separator(self):
        return "├" + "─" * (self.width - 2) + "┤"


fmt = TableFormatter()


class Range(object):
    """Class to keep a min/max of a value around"""

    def __init__(self):
        self.min = float("inf")
        self.max = float("-inf")

    def update(self, value):
        self.min = min(self.min, value)
        self.max = max(self.max, value)


class Touch(object):
    """A single data point of a sequence (i.e. one event frame)"""

    def __init__(self, pressure=None):
        self.pressure = pressure


class TouchSequence(object):
    """A touch sequence from beginning to end"""

    def __init__(self, device, tracking_id):
        self.device = device
        self.tracking_id = tracking_id
        self.points = []

        self.is_active = True

        self.is_down = False
        self.was_down = False
        self.is_palm = False
        self.was_palm = False
        self.is_thumb = False
        self.was_thumb = False

        self.prange = Range()

    def append(self, touch):
        """Add a Touch to the sequence"""
        self.points.append(touch)
        self.prange.update(touch.pressure)

        if touch.pressure < self.device.up:
            self.is_down = False
        elif touch.pressure > self.device.down:
            self.is_down = True
            self.was_down = True

        self.is_palm = touch.pressure > self.device.palm
        if self.is_palm:
            self.was_palm = True

        self.is_thumb = touch.pressure > self.device.thumb
        if self.is_thumb:
            self.was_thumb = True

    def finalize(self):
        """Mark the TouchSequence as complete (finger is up)"""
        self.is_active = False

    def avg(self):
        """Average pressure value of this sequence"""
        return int(sum([p.pressure for p in self.points]) / len(self.points))

    def median(self):
        """Median pressure value of this sequence"""
        ps = sorted([p.pressure for p in self.points])
        idx = int(len(self.points) / 2)
        return ps[idx]

    def __str__(self):
        return self._str_state() if self.is_active else self._str_summary()

    def _str_summary(self):
        if not self.points:
            return fmt.values(
                [
                    self.tracking_id,
                    False,
                    False,
                    False,
                    False,
                    "No pressure values recorded",
                ]
            )

        s = fmt.values(
            [
                self.tracking_id,
                self.was_down,
                True,
                self.was_palm,
                self.was_thumb,
                self.prange.min,
                self.prange.max,
                0,
                self.avg(),
                self.median(),
            ]
        )

        return s

    def _str_state(self):
        s = fmt.values(
            [
                self.tracking_id,
                self.is_down,
                not self.is_down,
                self.is_palm,
                self.is_thumb,
                self.prange.min,
                self.prange.max,
                self.points[-1].pressure,
            ]
        )
        return s


class InvalidDeviceError(Exception):
    pass


class Device(libevdev.Device):
    def __init__(self, path):
        if path is None:
            self.path = self.find_touchpad_device()
        else:
            self.path = path

        fd = open(self.path, "rb")
        flags = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        super().__init__(fd)

        print("Using {}: {}\n".format(self.name, self.path))

        self.has_mt_pressure = True
        absinfo = self.absinfo[libevdev.EV_ABS.ABS_MT_PRESSURE]
        if absinfo is None:
            absinfo = self.absinfo[libevdev.EV_ABS.ABS_PRESSURE]
            self.has_mt_pressure = False
            if absinfo is None:
                raise InvalidDeviceError(
                    "Device does not have ABS_PRESSURE or ABS_MT_PRESSURE"
                )

        prange = absinfo.maximum - absinfo.minimum

        # libinput defaults
        self.down = int(absinfo.minimum + 0.12 * prange)
        self.up = int(absinfo.minimum + 0.10 * prange)
        self.palm = 130  # the libinput default
        self.thumb = absinfo.maximum

        self._init_thresholds_from_quirks()
        self.sequences = []

    def find_touchpad_device(self):
        context = pyudev.Context()
        for device in context.list_devices(subsystem="input"):
            if not device.get("ID_INPUT_TOUCHPAD", 0):
                continue

            if not device.device_node or not device.device_node.startswith(
                "/dev/input/event"
            ):
                continue

            return device.device_node
        print("Unable to find a touchpad device.", file=sys.stderr)
        sys.exit(1)

    def _init_thresholds_from_quirks(self):
        command = ["libinput", "quirks", "list", self.path]
        cmd = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if cmd.returncode != 0:
            print(
                "Error querying quirks: {}".format(cmd.stderr.decode("utf-8")),
                file=sys.stderr,
            )
            return

        stdout = cmd.stdout.decode("utf-8")
        quirks = [q.split("=") for q in stdout.split("\n")]

        for q in quirks:
            if q[0] == "AttrPalmPressureThreshold":
                self.palm = int(q[1])
            elif q[0] == "AttrPressureRange":
                self.down, self.up = colon_tuple(q[1])
            elif q[0] == "AttrThumbPressureThreshold":
                self.thumb = int(q[1])

    def start_new_sequence(self, tracking_id):
        self.sequences.append(TouchSequence(self, tracking_id))

    def current_sequence(self):
        return self.sequences[-1]


def handle_key(device, event):
    tapcodes = [
        libevdev.EV_KEY.BTN_TOOL_DOUBLETAP,
        libevdev.EV_KEY.BTN_TOOL_TRIPLETAP,
        libevdev.EV_KEY.BTN_TOOL_QUADTAP,
        libevdev.EV_KEY.BTN_TOOL_QUINTTAP,
    ]
    if event.code in tapcodes and event.value > 0:
        try:
            if handle_key.warned:
                return
        except AttributeError:
            handle_key.warned = True
            print(
                "\r\033[2KThis tool cannot handle multiple fingers, "
                "output will be invalid",
            )


def handle_abs(device, event):
    if event.matches(libevdev.EV_ABS.ABS_MT_TRACKING_ID):
        if event.value > -1:
            device.start_new_sequence(event.value)
        else:
            try:
                s = device.current_sequence()
                s.finalize()
                print("\r\033[2K{}".format(s))
            except IndexError:
                # If the finger was down at startup
                pass
    elif event.matches(libevdev.EV_ABS.ABS_MT_PRESSURE) or (
        event.matches(libevdev.EV_ABS.ABS_PRESSURE) and not device.has_mt_pressure
    ):
        try:
            s = device.current_sequence()
            s.append(Touch(pressure=event.value))
            print("\r\033[2K{}".format(s))
        except IndexError:
            # If the finger was down at startup
            pass


def handle_event(device, event):
    if event.matches(libevdev.EV_ABS):
        handle_abs(device, event)
    elif event.matches(libevdev.EV_KEY):
        handle_key(device, event)


def loop(device):
    print("This is an interactive tool")
    print()
    print("Place a single finger on the touchpad to measure pressure values.")
    print("Check that:")
    print("- touches subjectively perceived as down are tagged as down")
    print("- touches with a thumb are tagged as thumb")
    print("- touches with a palm are tagged as palm")
    print()
    print("If the touch states do not match the interaction, re-run")
    print("with --touch-thresholds=down:up using observed pressure values.")
    print("See --help for more options.")
    print()
    print("Interactive keys:")
    print("  q/a - decrease/increase down threshold")
    print("  w/s - decrease/increase up threshold")
    print("  e/d - decrease/increase palm threshold")
    print("  r/f - decrease/increase thumb threshold")
    print()
    print("Press Ctrl+C to exit")
    print()

    headers = fmt.headers(
        ["Touch", "down", "up", "palm", "thumb", "min", "max", "p", "avg", "median"]
    )

    def print_thresholds():
        threshold_line = fmt.values(
            ["Thresh", device.down, device.up, device.palm, device.thumb]
        )
        print(f"\r{threshold_line}\r", end="", flush=True)

    print(fmt.header())
    print(headers)
    print(fmt.separator())
    print_thresholds()

    tty_settings = termios.tcgetattr(sys.stdin)
    try:
        tty.setcbreak(sys.stdin.fileno())
        while True:
            if select.select([sys.stdin], [], [], 0)[0]:
                key = sys.stdin.read(1)
                if key in "qawsedrf":
                    if key == "q":
                        device.down += 1
                    elif key == "a":
                        device.down = max(0, device.down - 1)
                        device.up = min(device.up, device.down)
                    elif key == "w":
                        device.up += 1
                        device.down = max(device.up, device.down)
                    elif key == "s":
                        device.up = max(0, device.up - 1)
                        device.down = max(device.up, device.down)
                    elif key == "e":
                        device.palm += 1
                    elif key == "d":
                        device.palm = max(0, device.palm - 1)
                    elif key == "r":
                        device.thumb += 1
                    elif key == "f":
                        device.thumb = max(0, device.thumb - 1)
                    print_thresholds()

            for event in device.events():
                handle_event(device, event)
                print_thresholds()
    finally:
        termios.tcsetattr(sys.stdin, termios.TCSADRAIN, tty_settings)
        print_thresholds()
        print()


def colon_tuple(string):
    try:
        ts = string.split(":")
        t = tuple([int(x) for x in ts])
        if len(t) == 2 and t[0] > t[1]:
            return t
    except:  # noqa
        pass

    msg = "{} is not in format N:M (N > M)".format(string)
    raise argparse.ArgumentTypeError(msg)


def main(args):
    parser = argparse.ArgumentParser(description="Measure touchpad pressure values")
    parser.add_argument(
        "path",
        metavar="/dev/input/event0",
        nargs="?",
        type=str,
        help="Path to device (optional)",
    )
    parser.add_argument(
        "--touch-thresholds",
        metavar="down:up",
        type=colon_tuple,
        help="Thresholds when a touch is logically down or up",
    )
    parser.add_argument(
        "--palm-threshold",
        metavar="t",
        type=int,
        help="Threshold when a touch is a palm",
    )
    parser.add_argument(
        "--thumb-threshold",
        metavar="t",
        type=int,
        help="Threshold when a touch is a thumb",
    )
    args = parser.parse_args()

    try:
        device = Device(args.path)

        if args.touch_thresholds is not None:
            device.down, device.up = args.touch_thresholds

        if args.palm_threshold is not None:
            device.palm = args.palm_threshold

        if args.thumb_threshold is not None:
            device.thumb = args.thumb_threshold

        loop(device)
    except KeyboardInterrupt:
        print("\r\033[2K{}".format(fmt.separator()))
        print()

    except (PermissionError, OSError):
        print("Error: failed to open device")
    except InvalidDeviceError as e:
        print(
            "This device does not have the capabilities for pressure-based touch detection."
        )
        print("Details: {}".format(e))


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2020 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


import sys
import argparse

try:
    import libevdev
    import pyudev
except ModuleNotFoundError as e:
    print("Error: {}".format(str(e)), file=sys.stderr)
    print(
        "One or more python modules are missing. Please install those "
        "modules and re-run this tool."
    )
    sys.exit(1)


class DeviceError(Exception):
    pass


class Point:
    def __init__(self, x=None, y=None):
        self.x = x
        self.y = y


class Touchpad(object):
    def __init__(self, evdev):
        x = evdev.absinfo[libevdev.EV_ABS.ABS_X]
        y = evdev.absinfo[libevdev.EV_ABS.ABS_Y]
        if not x or not y:
            raise DeviceError("Device does not have an x or axis")

        if not x.resolution or not y.resolution:
            print("Device does not have resolutions.", file=sys.stderr)
            x.resolution = 1
            y.resolution = 1

        self.xrange = x.maximum - x.minimum
        self.yrange = y.maximum - y.minimum
        self.width = self.xrange / x.resolution
        self.height = self.yrange / y.resolution

        self._x = x
        self._y = y

        # We try to make the touchpad at least look proportional. The
        # terminal character space is (guesswork) ca 2.3 times as high as
        # wide.
        self.columns = 30
        self.rows = int(
            self.columns
            * (self.yrange // y.resolution)
            // (self.xrange // x.resolution)
            / 2.3
        )
        self.pos = Point(0, 0)
        self.min = Point()
        self.max = Point()

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @x.setter
    def x(self, x):
        self._x.minimum = min(self.x.minimum, x)
        self._x.maximum = max(self.x.maximum, x)
        self.min.x = min(x, self.min.x or 0xFFFFFFFF)
        self.max.x = max(x, self.max.x or -0xFFFFFFFF)
        # we calculate the position based on the original range.
        # this means on devices with a narrower range than advertised, not
        # all corners may be reachable in the touchpad drawing.
        self.pos.x = min(0.99, (x - self._x.minimum) / self.xrange)

    @y.setter
    def y(self, y):
        self._y.minimum = min(self.y.minimum, y)
        self._y.maximum = max(self.y.maximum, y)
        self.min.y = min(y, self.min.y or 0xFFFFFFFF)
        self.max.y = max(y, self.max.y or -0xFFFFFFFF)
        # we calculate the position based on the original range.
        # this means on devices with a narrower range than advertised, not
        # all corners may be reachable in the touchpad drawing.
        self.pos.y = min(0.99, (y - self._y.minimum) / self.yrange)

    def update_from_data(self):
        if None in [self.min.x, self.min.y, self.max.x, self.max.y]:
            raise DeviceError("Insufficient data to continue")
        self._x.minimum = self.min.x
        self._x.maximum = self.max.x
        self._y.minimum = self.min.y
        self._y.maximum = self.max.y

    def draw(self):
        print(
            "Detected axis range: x [{:4d}..{:4d}], y [{:4d}..{:4d}]".format(
                self.min.x if self.min.x is not None else 0,
                self.max.x if self.max.x is not None else 0,
                self.min.y if self.min.y is not None else 0,
                self.max.y if self.max.y is not None else 0,
            )
        )

        print()
        print("Move one finger along all edges of the touchpad".center(self.columns))
        print("until the detected axis range stops changing.".center(self.columns))

        top = int(self.pos.y * self.rows)

        print("+{}+".format("".ljust(self.columns, "-")))
        for row in range(0, top):
            print("|{}|".format("".ljust(self.columns)))

        left = int(self.pos.x * self.columns)
        right = max(0, self.columns - 1 - left)
        print("|{}{}{}|".format("".ljust(left), "O", "".ljust(right)))

        for row in range(top + 1, self.rows):
            print("|{}|".format("".ljust(self.columns)))

        print("+{}+".format("".ljust(self.columns, "-")))

        print("Press Ctrl+C to stop".center(self.columns))

        print("\033[{}A".format(self.rows + 8), flush=True)

        self.rows_printed = self.rows + 8

    def erase(self):
        # Erase all previous lines so we're not left with rubbish
        for row in range(self.rows_printed):
            print("\033[K")
        print("\033[{}A".format(self.rows_printed))


def dimension(string):
    try:
        ts = string.split("x")
        t = tuple([int(x) for x in ts])
        if len(t) == 2:
            return t
    except:  # noqa
        pass

    msg = "{} is not in format WxH".format(string)
    raise argparse.ArgumentTypeError(msg)


def between(v1, v2, deviation):
    return v1 - deviation < v2 < v1 + deviation


def dmi_modalias_match(modalias):
    modalias = modalias.split(":")
    dmi = {"svn": None, "pvr": None, "pn": None}
    for m in modalias:
        for key in dmi:
            if m.startswith(key):
                dmi[key] = m[len(key) :]

    # Based on the current 60-evdev.hwdb, Lenovo uses pvr and everyone else
    # uses pn to provide a human-identifiable match
    if dmi["svn"] == "LENOVO":
        return "dmi:*svn{}:*pvr{}*".format(dmi["svn"], dmi["pvr"])
    else:
        return "dmi:*svn{}:*pn{}*".format(dmi["svn"], dmi["pn"])


def main(args):
    parser = argparse.ArgumentParser(description="Measure the touchpad size")
    parser.add_argument(
        "size",
        metavar="WxH",
        type=dimension,
        help="Touchpad size (width by height) in mm",
    )
    parser.add_argument(
        "path",
        metavar="/dev/input/event0",
        nargs="?",
        type=str,
        help="Path to device (optional)",
    )
    context = pyudev.Context()

    args = parser.parse_args()
    if not args.path:
        for device in context.list_devices(subsystem="input"):
            if device.get("ID_INPUT_TOUCHPAD", 0) and (
                device.device_node or ""
            ).startswith("/dev/input/event"):
                args.path = device.device_node
                name = "unknown"
                parent = device
                while parent is not None:
                    n = parent.get("NAME", None)
                    if n:
                        name = n
                        break
                    parent = parent.parent

                print("Using {}: {}".format(name, device.device_node))
                break
        else:
            print("Unable to find a touchpad device.", file=sys.stderr)
            return 1

    dev = pyudev.Devices.from_device_file(context, args.path)
    overrides = [p for p in dev.properties if p.startswith("EVDEV_ABS")]
    if overrides:
        print()
        print("********************************************************************")
        print("WARNING: axis overrides already in place for this device:")
        for prop in overrides:
            print("  {}={}".format(prop, dev.properties[prop]))
        print("The systemd hwdb already overrides the axis ranges and/or resolution.")
        print("This tool is not needed unless you want to verify the axis overrides.")
        print("********************************************************************")
        print()

    try:
        fd = open(args.path, "rb")
        evdev = libevdev.Device(fd)
        touchpad = Touchpad(evdev)
        print(
            "Kernel specified touchpad size: {:.1f}x{:.1f}mm".format(
                touchpad.width, touchpad.height
            )
        )
        print("User specified touchpad size:   {:.1f}x{:.1f}mm".format(*args.size))

        print()
        print(
            "Kernel axis range:   x [{:4d}..{:4d}], y [{:4d}..{:4d}]".format(
                touchpad.x.minimum,
                touchpad.x.maximum,
                touchpad.y.minimum,
                touchpad.y.maximum,
            )
        )

        print("Put your finger on the touchpad to start\033[1A")

        try:
            touchpad.draw()
            while True:
                for event in evdev.events():
                    if event.matches(libevdev.EV_ABS.ABS_X):
                        touchpad.x = event.value
                    elif event.matches(libevdev.EV_ABS.ABS_Y):
                        touchpad.y = event.value
                    elif event.matches(libevdev.EV_SYN.SYN_REPORT):
                        touchpad.draw()
        except KeyboardInterrupt:
            touchpad.erase()
            touchpad.update_from_data()

        print(
            "Detected axis range: x [{:4d}..{:4d}], y [{:4d}..{:4d}]".format(
                touchpad.x.minimum,
                touchpad.x.maximum,
                touchpad.y.minimum,
                touchpad.y.maximum,
            )
        )

        touchpad.x.resolution = round(
            (touchpad.x.maximum - touchpad.x.minimum) / args.size[0]
        )
        touchpad.y.resolution = round(
            (touchpad.y.maximum - touchpad.y.minimum) / args.size[1]
        )

        print(
            "Resolutions calculated based on user-specified size: x {}, y {} units/mm".format(
                touchpad.x.resolution, touchpad.y.resolution
            )
        )

        # If both x/y are within some acceptable deviation, we skip the axis
        # overrides and only override the resolution
        xorig = evdev.absinfo[libevdev.EV_ABS.ABS_X]
        yorig = evdev.absinfo[libevdev.EV_ABS.ABS_Y]
        deviation = 1.5 * touchpad.x.resolution  # 1.5 mm rounding on each side
        skip = between(xorig.minimum, touchpad.x.minimum, deviation)
        skip = skip and between(xorig.maximum, touchpad.x.maximum, deviation)
        deviation = 1.5 * touchpad.y.resolution  # 1.5 mm rounding on each side
        skip = skip and between(yorig.minimum, touchpad.y.minimum, deviation)
        skip = skip and between(yorig.maximum, touchpad.y.maximum, deviation)

        if skip:
            print()
            print(
                "Note: Axis ranges within acceptable deviation, skipping min/max override"
            )
            print()

        print()
        print("Suggested hwdb entry:")

        use_dmi = evdev.id["bustype"] not in [0x03, 0x05]  # USB, Bluetooth
        if use_dmi:
            modalias = open("/sys/class/dmi/id/modalias").read().strip()
            print(
                "Note: the dmi modalias match is a guess based on your machine's modalias:"
            )
            print(" ", modalias)
            print(
                "Please verify that this is the most sensible match and adjust if necessary."
            )

        print("-8<--------------------------")
        print("# Laptop model description (e.g. Lenovo X1 Carbon 5th)")
        if use_dmi:
            print("evdev:name:{}:{}*".format(evdev.name, dmi_modalias_match(modalias)))
        else:
            print(
                "evdev:input:b{:04X}v{:04X}p{:04X}*".format(
                    evdev.id["bustype"], evdev.id["vendor"], evdev.id["product"]
                )
            )
        print(
            " EVDEV_ABS_00={}:{}:{}".format(
                touchpad.x.minimum if not skip else "",
                touchpad.x.maximum if not skip else "",
                touchpad.x.resolution,
            )
        )
        print(
            " EVDEV_ABS_01={}:{}:{}".format(
                touchpad.y.minimum if not skip else "",
                touchpad.y.maximum if not skip else "",
                touchpad.y.resolution,
            )
        )
        if evdev.absinfo[libevdev.EV_ABS.ABS_MT_POSITION_X]:
            print(
                " EVDEV_ABS_35={}:{}:{}".format(
                    touchpad.x.minimum if not skip else "",
                    touchpad.x.maximum if not skip else "",
                    touchpad.x.resolution,
                )
            )
            print(
                " EVDEV_ABS_36={}:{}:{}".format(
                    touchpad.y.minimum if not skip else "",
                    touchpad.y.maximum if not skip else "",
                    touchpad.y.resolution,
                )
            )
        print("-8<--------------------------")
        print(
            "Instructions on what to do with this snippet are in /usr/lib/udev/hwdb.d/60-evdev.hwdb"
        )
    except DeviceError as e:
        print("Error: {}".format(e), file=sys.stderr)
        return 1
    except PermissionError:
        print("Unable to open device. Please run me as root", file=sys.stderr)
        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



#!/usr/bin/env python3
# -*- coding: utf-8
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the 'Software'),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#
#
# Measures the relative motion between touch events (based on slots)
#
# Input is a libinput record yaml file

from dataclasses import dataclass, field, replace
from enum import Enum

import argparse
import math
import sys
import yaml
import libevdev


COLOR_RESET = "\x1b[0m"
COLOR_RED = "\x1b[6;31m"
COLOR_BLUE = "\x1b[6;34m"
COLOR_GREEN = "\x1b[6;32m"


class SlotFormatter:
    def __init__(
        self,
        is_absolute=False,
        resolution=None,
        threshold=None,
        ignore_below=None,
        show_distance=False,
        pressure_thresholds=(0, 0),
    ):
        self.threshold = threshold
        self.ignore_below = ignore_below
        self.resolution = resolution
        self.is_absolute = is_absolute
        self.show_distance = show_distance
        self.pressure_thresholds = pressure_thresholds
        self.slots = []
        self.have_data = False
        self.filtered = False

        self.width = 35 if show_distance else 16

    def __str__(self):
        return " | ".join(self.slots)

    def format_slot(self, slot):
        if slot.state == SlotState.BEGIN:
            self.slots.append("+++++++".center(self.width))
            self.have_data = True
        elif slot.state == SlotState.END:
            self.slots.append("-------".center(self.width))
            self.have_data = True
        elif slot.state == SlotState.NONE:
            self.slots.append(("*" * (self.width - 2)).center(self.width))
        elif not slot.dirty:
            self.slots.append(" ".center(self.width))
        else:
            if self.resolution is not None:
                delta = Point(
                    slot.delta.x / self.resolution[0],
                    slot.delta.y / self.resolution[1],
                )
                distx = abs(slot.position.x - slot.origin.x)
                disty = abs(slot.position.y - slot.origin.y)
                distance = Point(
                    distx / self.resolution[0],
                    disty / self.resolution[1],
                )
            else:
                delta = Point(slot.delta.x, slot.delta.y)
                distance = Point(
                    abs(slot.position.x - slot.origin.x),
                    abs(slot.position.y - slot.origin.y),
                )

            if delta.x != 0 and delta.y != 0:
                t = math.atan2(delta.x, delta.y)
                t += math.pi  # in [0, 2pi] range now

                if t == 0:
                    t = 0.01
                else:
                    t = t * 180.0 / math.pi

                directions = ["↖↑", "↖←", "↙←", "↙↓", "↓↘", "→↘", "→↗", "↑↗"]
                direction = directions[int(t / 45)]
            elif delta.y == 0:
                if delta.x < 0:
                    direction = "←←"
                else:
                    direction = "→→"
            else:
                if delta.y < 0:
                    direction = "↑↑"
                else:
                    direction = "↓↓"

            color = COLOR_RESET
            reset = COLOR_RESET
            if not self.is_absolute:
                if (
                    self.pressure_thresholds[1] > 0
                    and slot.pressure > self.pressure_thresholds[1]
                ):
                    color = COLOR_GREEN
                    reset = COLOR_RESET
                elif (
                    self.pressure_thresholds[0] > 0
                    and slot.pressure > self.pressure_thresholds[0]
                ):
                    color = COLOR_BLUE
                    reset = COLOR_RESET

                if self.ignore_below is not None or self.threshold is not None:
                    dist = math.hypot(delta.x, delta.y)
                    if self.ignore_below is not None and dist < self.ignore_below:
                        self.slots.append(" ".center(self.width))
                        self.filtered = True
                        return
                    if self.threshold is not None and dist >= self.threshold:
                        color = COLOR_RED
                        reset = COLOR_RESET

                if isinstance(delta.x, int) and isinstance(delta.y, int):
                    coords = f"{delta.x:+4d}/{delta.y:+4d}"
                else:
                    coords = f"{delta.x:+3.2f}/{delta.y:+03.2f}"

                if self.show_distance:
                    hypot = math.hypot(distance.x, distance.y)
                    distance = (
                        f"dist: ({distance.x:3.1f}/{distance.y:3.1f}, {hypot:3.1f})"
                    )
                else:
                    distance = ""

                components = [
                    f"{direction}",
                    f"{color}",
                    coords,
                    distance,
                    f"{reset}",
                ]

                string = " ".join(c for c in components if c)
            else:
                x, y = slot.position.x, slot.position.y
                string = "{} {}{:4d}/{:4d}{}".format(direction, color, x, y, reset)
            self.have_data = True
            self.slots.append(string.ljust(self.width + len(color) + len(reset)))


class SlotState(Enum):
    NONE = 0
    BEGIN = 1
    UPDATE = 2
    END = 3


@dataclass
class Point:
    x: float = 0.0
    y: float = 0.0


@dataclass
class Slot:
    index: int
    state: SlotState = SlotState.NONE
    position: Point = field(default_factory=Point)
    delta: Point = field(default_factory=Point)
    origin: Point = field(default_factory=Point)
    pressure: int = 0
    used: bool = False
    dirty: bool = False


def main(argv):
    global COLOR_RESET
    global COLOR_RED
    global COLOR_BLUE
    global COLOR_GREEN

    slots = []
    xres, yres = 1, 1

    parser = argparse.ArgumentParser(
        description="Measure delta between event frames for each slot"
    )
    parser.add_argument(
        "--use-mm", action="store_true", help="Use mm instead of device deltas"
    )
    parser.add_argument(
        "--show-distance",
        action="store_true",
        help="Show the absolute distance relative to the first position",
    )
    parser.add_argument(
        "--use-st",
        action="store_true",
        help="Use ABS_X/ABS_Y instead of ABS_MT_POSITION_X/Y",
    )
    parser.add_argument(
        "--use-absolute",
        action="store_true",
        help="Use absolute coordinates, not deltas",
    )
    parser.add_argument(
        "path", metavar="recording", nargs=1, help="Path to libinput-record YAML file"
    )
    parser.add_argument(
        "--threshold",
        type=float,
        default=None,
        help="Mark any delta above this threshold",
    )
    parser.add_argument(
        "--ignore-below",
        type=float,
        default=None,
        help="Ignore any delta below this threshold",
    )
    parser.add_argument(
        "--pressure-min",
        type=int,
        default=0,
        help="Highlight touches above this pressure minimum",
    )
    parser.add_argument(
        "--pressure-max",
        type=int,
        default=0,
        help="Highlight touches below this pressure maximum",
    )
    args = parser.parse_args()

    if not sys.stdout.isatty():
        COLOR_RESET = ""
        COLOR_RED = ""
        COLOR_GREEN = ""
        COLOR_BLUE = ""

    yml = yaml.safe_load(open(args.path[0]))
    device = yml["devices"][0]
    absinfo = device["evdev"]["absinfo"]
    try:
        nslots = absinfo[libevdev.EV_ABS.ABS_MT_SLOT.value][1] + 1
    except KeyError:
        args.use_st = True

    if args.use_st:
        nslots = 1

    slots = [Slot(i) for i in range(0, nslots)]
    slots[0].used = True

    if args.use_mm:
        xres = 1.0 * absinfo[libevdev.EV_ABS.ABS_X.value][4]
        yres = 1.0 * absinfo[libevdev.EV_ABS.ABS_Y.value][4]
        if not xres or not yres:
            print("Error: device doesn't have a resolution, cannot use mm")
            sys.exit(1)

    if args.use_st:
        print("Warning: slot coordinates on FINGER/DOUBLETAP change may be incorrect")
        slots[0].used = True

    slot = 0
    last_time = None
    tool_bits = {
        libevdev.EV_KEY.BTN_TOUCH: 0,
        libevdev.EV_KEY.BTN_TOOL_DOUBLETAP: 0,
        libevdev.EV_KEY.BTN_TOOL_TRIPLETAP: 0,
        libevdev.EV_KEY.BTN_TOOL_QUADTAP: 0,
        libevdev.EV_KEY.BTN_TOOL_QUINTTAP: 0,
    }
    btn_state = {
        libevdev.EV_KEY.BTN_LEFT: 0,
        libevdev.EV_KEY.BTN_MIDDLE: 0,
        libevdev.EV_KEY.BTN_RIGHT: 0,
    }

    nskipped_lines = 0

    for event in device["events"]:
        for evdev in event["evdev"]:
            s = slots[slot]
            e = libevdev.InputEvent(
                code=libevdev.evbit(evdev[2], evdev[3]),
                value=evdev[4],
                sec=evdev[0],
                usec=evdev[1],
            )

            if e.code in tool_bits:
                tool_bits[e.code] = e.value
            if e.code in btn_state:
                btn_state[e.code] = e.value

            if args.use_st:
                # Note: this relies on the EV_KEY events to come in before the
                # x/y events, otherwise the last/first event in each slot will
                # be wrong.
                if (
                    e.code == libevdev.EV_KEY.BTN_TOOL_FINGER
                    or e.code == libevdev.EV_KEY.BTN_TOOL_PEN
                ):
                    slot = 0
                    s = slots[slot]
                    s.dirty = True
                    if e.value:
                        s.state = SlotState.BEGIN
                    else:
                        s.state = SlotState.END
                elif e.code == libevdev.EV_KEY.BTN_TOOL_DOUBLETAP:
                    if len(slots) > 1:
                        slot = 1
                    s = slots[slot]
                    s.dirty = True
                    if e.value:
                        s.state = SlotState.BEGIN
                    else:
                        s.state = SlotState.END
                elif e.code == libevdev.EV_ABS.ABS_PRESSURE:
                    s.pressure = e.value
            else:
                if e.code == libevdev.EV_ABS.ABS_MT_SLOT:
                    slot = e.value
                    s = slots[slot]
                    s.dirty = True
                    # bcm5974 cycles through slot numbers, so let's say all below
                    # our current slot number was used
                    for sl in slots[: slot + 1]:
                        sl.used = True
                elif e.code == libevdev.EV_ABS.ABS_MT_TRACKING_ID:
                    if e.value == -1:
                        s.state = SlotState.END
                    else:
                        s.state = SlotState.BEGIN
                        s.delta.x, s.delta.y = 0, 0
                    s.dirty = True
                elif e.code == libevdev.EV_ABS.ABS_MT_PRESSURE:
                    s.pressure = e.value

            if args.use_st:
                axes = [libevdev.EV_ABS.ABS_X, libevdev.EV_ABS.ABS_Y]
            else:
                axes = [
                    libevdev.EV_ABS.ABS_MT_POSITION_X,
                    libevdev.EV_ABS.ABS_MT_POSITION_Y,
                ]

            if e.code in axes:
                s.dirty = True

                # If recording started after touch down
                if s.state == SlotState.NONE:
                    s.state = SlotState.BEGIN
                    s.delta = Point(0, 0)

                if e.code in [
                    libevdev.EV_ABS.ABS_X,
                    libevdev.EV_ABS.ABS_MT_POSITION_X,
                ]:
                    if s.state == SlotState.UPDATE:
                        s.delta.x = e.value - s.position.x
                    s.position.x = e.value
                elif e.code in [
                    libevdev.EV_ABS.ABS_Y,
                    libevdev.EV_ABS.ABS_MT_POSITION_Y,
                ]:
                    if s.state == SlotState.UPDATE:
                        s.delta.y = e.value - s.position.y
                    s.position.y = e.value
                else:
                    assert False, f"Invalid axis {e.code}"

            if e.code == libevdev.EV_SYN.SYN_REPORT:
                if last_time is None:
                    last_time = e.sec * 1000000 + e.usec
                    tdelta = 0
                else:
                    t = e.sec * 1000000 + e.usec
                    tdelta = int((t - last_time) / 1000)  # ms
                    last_time = t

                tools = [
                    (libevdev.EV_KEY.BTN_TOOL_QUINTTAP, "QIN"),
                    (libevdev.EV_KEY.BTN_TOOL_QUADTAP, "QAD"),
                    (libevdev.EV_KEY.BTN_TOOL_TRIPLETAP, "TRI"),
                    (libevdev.EV_KEY.BTN_TOOL_DOUBLETAP, "DBL"),
                    (libevdev.EV_KEY.BTN_TOUCH, "TOU"),
                ]

                for bit, string in tools:
                    if tool_bits[bit]:
                        tool_state = string
                        break
                else:
                    tool_state = "   "

                buttons = [
                    (libevdev.EV_KEY.BTN_LEFT, "L"),
                    (libevdev.EV_KEY.BTN_MIDDLE, "M"),
                    (libevdev.EV_KEY.BTN_RIGHT, "R"),
                ]

                button_state = (
                    "".join([string for bit, string in buttons if btn_state[bit]])
                    or "."
                )

                fmt = SlotFormatter(
                    is_absolute=args.use_absolute,
                    resolution=(xres, yres) if args.use_mm else None,
                    threshold=args.threshold,
                    ignore_below=args.ignore_below,
                    show_distance=args.show_distance,
                    pressure_thresholds=(args.pressure_min, args.pressure_max),
                )
                for sl in [s for s in slots if s.used]:
                    fmt.format_slot(sl)

                    sl.dirty = False
                    sl.delta.x, sl.delta.y = 0, 0
                    if sl.state == SlotState.BEGIN:
                        sl.origin = replace(sl.position)
                        sl.state = SlotState.UPDATE
                    elif sl.state == SlotState.END:
                        sl.state = SlotState.NONE

                if fmt.have_data:
                    if nskipped_lines > 0:
                        print("")
                        nskipped_lines = 0
                    print(
                        "{:2d}.{:06d} {:+5d}ms {} {} {}".format(
                            e.sec, e.usec, tdelta, tool_state, button_state, fmt
                        )
                    )
                elif fmt.filtered:
                    nskipped_lines += 1
                    print(
                        "\r",
                        " " * 21,
                        "... {} below threshold".format(nskipped_lines),
                        flush=True,
                        end="",
                    )


if __name__ == "__main__":
    try:
        main(sys.argv)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2018 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import os
import sys
import time
import math
import multiprocessing
import argparse
from pathlib import Path
from tempfile import NamedTemporaryFile

try:
    import libevdev
    import yaml
    import pyudev
except ModuleNotFoundError as e:
    print("Error: {}".format(e), file=sys.stderr)
    print(
        "One or more python modules are missing. Please install those "
        "modules and re-run this tool."
    )
    sys.exit(1)


SUPPORTED_FILE_VERSION = 1


def error(msg, **kwargs):
    print(msg, **kwargs, file=sys.stderr)


class YamlException(Exception):
    pass


def fetch(yaml, key):
    """Helper function to avoid confusing a YAML error with a
    normal KeyError bug"""
    try:
        return yaml[key]
    except KeyError:
        raise YamlException("Failed to get '{}' from recording.".format(key))


def check_udev_properties(yaml_data, uinput):
    """
    Compare the properties our new uinput device has with the ones from the
    recording and ring the alarm bell if one of them is off.
    """
    yaml_udev_section = fetch(yaml_data, "udev")
    yaml_udev_props = fetch(yaml_udev_section, "properties")

    ignore = ["LIBINPUT_DEVICE_GROUP", "DRIVER"]
    yaml_props = {
        k: v
        for (k, v) in [prop.split("=", maxsplit=1) for prop in yaml_udev_props]
        if k not in ignore
    }

    # give udev some time to catch up
    time.sleep(0.2)
    context = pyudev.Context()
    udev_device = pyudev.Devices.from_device_file(context, uinput.devnode)
    for name, value in udev_device.properties.items():
        if name in yaml_props:
            if yaml_props[name] != value:
                error(
                    f"Warning: udev property mismatch: recording has {name}={yaml_props[name]}, device has {name}={value}"
                )
            del yaml_props[name]
        else:
            # The list of properties we add to the recording, see libinput-record.c
            prefixes = (
                "ID_INPUT",
                "LIBINPUT",
                "EVDEV_ABS",
                "MOUSE_DPI",
                "POINTINGSTICK_",
            )
            for prefix in prefixes:
                if name.startswith(prefix):
                    error(f"Warning: unexpected property: {name}={value}")

    # the ones we found above were removed from the dict
    for name, value in yaml_props.items():
        error(f"Warning: device is missing recorded udev property: {name}={value}")


def create(device):
    evdev = fetch(device, "evdev")

    d = libevdev.Device()
    d.name = fetch(evdev, "name")

    ids = fetch(evdev, "id")
    if len(ids) != 4:
        raise YamlException("Invalid ID format: {}".format(ids))
    d.id = dict(zip(["bustype", "vendor", "product", "version"], ids))

    codes = fetch(evdev, "codes")
    for evtype, evcodes in codes.items():
        for code in evcodes:
            data = None
            if evtype == libevdev.EV_ABS.value:
                values = fetch(evdev, "absinfo")[code]
                absinfo = libevdev.InputAbsInfo(
                    minimum=values[0],
                    maximum=values[1],
                    fuzz=values[2],
                    flat=values[3],
                    resolution=values[4],
                )
                data = absinfo
            elif evtype == libevdev.EV_REP.value:
                if code == libevdev.EV_REP.REP_DELAY.value:
                    data = 500
                elif code == libevdev.EV_REP.REP_PERIOD.value:
                    data = 20
            d.enable(libevdev.evbit(evtype, code), data=data)

    properties = fetch(evdev, "properties")
    for prop in properties:
        d.enable(libevdev.propbit(prop))

    uinput = d.create_uinput_device()

    check_udev_properties(device, uinput)

    return uinput


def print_events(devnode, indent, evs):
    devnode = os.path.basename(devnode)
    for e in evs:
        if e.type != libevdev.EV_SYN:
            print(
                "{}: {}{:-6d}.{:06d} {} / {:<20s} {:6d}".format(
                    devnode,
                    " " * (indent * 8),
                    e.sec,
                    e.usec,
                    e.type.name,
                    e.code.name,
                    e.value,
                )
            )
        if e.type == libevdev.EV_SYN:
            print(
                "{}: {}----------------- SYN_REPORT ({}) -----------------".format(
                    devnode, " " * (indent * 8), e.value
                )
            )


def collect_events(frame):
    evs = []
    events_skipped = False
    for sec, usec, evtype, evcode, value in frame:
        if evtype == libevdev.EV_KEY.value and value == 2:  # key repeat
            events_skipped = True
            continue

        e = libevdev.InputEvent(
            libevdev.evbit(evtype, evcode), value=value, sec=sec, usec=usec
        )
        evs.append(e)

    # If we skipped some events and now all we have left is the
    # SYN_REPORTs, we drop the SYN_REPORTs as well.
    if events_skipped and all(e for e in evs if e.matches(libevdev.EV_SYN.SYN_REPORT)):
        return []
    else:
        return evs


def replay(device, verbose):
    events = fetch(device, "events")
    if events is None:
        return
    uinput = device["__uinput"]

    # The first event may have a nonzero offset but we want to replay
    # immediately regardless. When replaying multiple devices, the first
    # offset is the offset from the first event on any device.
    offset = time.time() - device["__first_event_offset"]

    if offset < 0:
        error("WARNING: event time offset is in the future, refusing to replay")
        return

    # each 'evdev' set contains one SYN_REPORT so we only need to check for
    # the time offset once per event
    for event in events:
        try:
            evdev = fetch(event, "evdev")
        except YamlException:
            continue

        evs = collect_events(evdev)
        if not evs:
            continue

        evtime = evs[0].sec + evs[0].usec / 1e6 + offset
        now = time.time()

        if evtime - now > 150 / 1e6:  # 150 µs error margin
            time.sleep(evtime - now - 150 / 1e6)

        uinput.send_events(evs)
        if verbose:
            print_events(uinput.devnode, device["__index"], evs)


def first_timestamp(device):
    events = fetch(device, "events")
    for e in events or []:
        try:
            evdev = fetch(e, "evdev")
            (sec, usec, *_) = evdev[0]
            return sec + usec / 1.0e6
        except YamlException:
            pass

    return None


def wrap(func, *args):
    try:
        func(*args)
    except KeyboardInterrupt:
        pass


def loop(args, recording):
    devices = fetch(recording, "devices")

    first_timestamps = tuple(
        filter(lambda x: x is not None, [first_timestamp(d) for d in devices])
    )
    # All devices need to start replaying at the same time, so let's find
    # the very first event and offset everything by that timestamp.
    toffset = min(first_timestamps or [math.inf])

    for idx, d in enumerate(devices):
        uinput = create(d)
        print("{}: {}".format(uinput.devnode, uinput.name))
        d["__uinput"] = uinput  # cheaper to hide it in the dict then work around it
        d["__index"] = idx
        d["__first_event_offset"] = toffset

    if not first_timestamps:
        input("No events in recording. Hit enter to quit")
        return

    while True:
        if args.replay_after >= 0:
            time.sleep(args.replay_after)
        else:
            input("Hit enter to start replaying")

        try:
            processes = [
                multiprocessing.Process(target=wrap, args=(replay, d, args.verbose))
                for d in devices
            ]

            for p in processes:
                p.start()

            for p in processes:
                p.join()

            if args.once:
                break
        except KeyboardInterrupt:
            print("Event replay interrupted, press Ctrl+C again to exit.")
            print("Note that the device may not be in a neutral state now.")


def create_device_quirk(device, quirks):
    # Where the device has a quirk, we match on name, vendor and product.
    # That's the best match we can assemble here from the info we have.
    evdev = fetch(device, "evdev")
    name = fetch(evdev, "name")
    id = fetch(evdev, "id")
    quirk = (
        "[libinput-replay {name}]\n"
        "MatchName={name}\n"
        "MatchVendor=0x{id[1]:04X}\n"
        "MatchProduct=0x{id[2]:04X}\n"
    ).format(name=name, id=id)
    quirk += "\n".join(quirks)
    return quirk


def setup_quirks(recording) -> Path | None:
    devices = fetch(recording, "devices")
    quirks = []
    for d in devices:
        qs = d.get("quirks") or []
        if not any(q.startswith("AttrIsVirtual=") for q in qs):
            try:
                is_virtual = d["udev"]["virtual"]
            except Exception:
                is_virtual = False
            qs.append(f"AttrIsVirtual={int(is_virtual)}")
        quirks.append(create_device_quirk(d, qs))
    if not quirks:
        return None

    runtime_dir = (
        Path(os.getenv("XDG_RUNTIME_DIR", f"/run/user/{os.geteuid()}")) / "libinput"
    )
    runtime_dir.mkdir(exist_ok=True, parents=True)
    with NamedTemporaryFile(
        mode="w+",
        dir=runtime_dir,
        suffix=".quirks",
        prefix="libinput-replay",
        delete=False,
    ) as fd:
        fd.write("# This file was generated by libinput replay\n")
        fd.write("# Unless libinput replay is running right now, remove this file.\n")
        fd.write("\n\n".join(quirks))

        return Path(fd.name)


def check_file(recording):
    version = fetch(recording, "version")
    if version != SUPPORTED_FILE_VERSION:
        raise YamlException(
            "Invalid file format: {}, expected {}".format(
                version, SUPPORTED_FILE_VERSION
            )
        )

    ndevices = fetch(recording, "ndevices")
    devices = fetch(recording, "devices")
    if ndevices != len(devices):
        error(
            "WARNING: truncated file, expected {} devices, got {}".format(
                ndevices, len(devices)
            )
        )


def main():
    multiprocessing.set_start_method("fork")

    parser = argparse.ArgumentParser(description="Replay a device recording")
    parser.add_argument(
        "recording",
        metavar="recorded-file.yaml",
        type=str,
        help="Path to device recording",
    )
    parser.add_argument(
        "--replay-after",
        type=int,
        default=-1,
        help="Automatically replay once after N seconds",
    )
    parser.add_argument(
        "--once",
        action="store_true",
        default=False,
        help="Stop and exit after one replay",
    )
    parser.add_argument("--verbose", action="store_true")
    args = parser.parse_args()

    quirks_file = None

    try:
        with open(args.recording) as f:
            y = yaml.safe_load(f)
            check_file(y)
            quirks_file = setup_quirks(y)
            loop(args, y)
    except KeyboardInterrupt:
        pass
    except (PermissionError, OSError) as e:
        error("Error: failed to open device: {}".format(e))
    except YamlException as e:
        error("Error: failed to parse recording: {}".format(e))
    finally:
        if quirks_file:
            quirks_file.unlink()
            try:
                quirks_file.parent.rmdir()
            except OSError as e:
                import errno

                if e.errno != errno.ENOTEMPTY:
                    raise e


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2017 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#

import sys
import argparse

try:
    import libevdev
    import textwrap
    import pyudev
except ModuleNotFoundError as e:
    print("Error: {}".format(e), file=sys.stderr)
    print(
        "One or more python modules are missing. Please install those "
        "modules and re-run this tool."
    )
    sys.exit(1)

print_dest = sys.stdout


def error(msg, **kwargs):
    print(msg, **kwargs, file=sys.stderr)


def msg(msg, **kwargs):
    print(msg, **kwargs, file=print_dest, flush=True)


def tv2us(sec, usec):
    return sec * 1000000 + usec


def us2ms(us):
    return int(us / 1000)


class Touch(object):
    def __init__(self, down):
        self._down = down
        self._up = down

    @property
    def up(self):
        return us2ms(self._up)

    @up.setter
    def up(self, up):
        assert up > self.down
        self._up = up

    @property
    def down(self):
        return us2ms(self._down)

    @property
    def tdelta(self):
        return self.up - self.down


class InvalidDeviceError(Exception):
    pass


class Device(libevdev.Device):
    def __init__(self, path):
        if path is None:
            self.path = self._find_touch_device()
        else:
            self.path = path
        fd = open(self.path, "rb")
        super().__init__(fd)

        print("Using {}: {}\n".format(self.name, self.path))

        if not self.has(libevdev.EV_KEY.BTN_TOUCH):
            raise InvalidDeviceError("device does not have BTN_TOUCH")

        self.touches = []
        self.warned = False

    def _find_touch_device(self):
        context = pyudev.Context()
        device_node = None
        for device in context.list_devices(subsystem="input"):
            if not device.device_node or not device.device_node.startswith(
                "/dev/input/event"
            ):
                continue

            # pick the touchpad by default, fallback to the first
            # touchscreen only when there is no touchpad
            if device.get("ID_INPUT_TOUCHPAD", 0):
                device_node = device.device_node
                break

            if device.get("ID_INPUT_TOUCHSCREEN", 0) and device_node is None:
                device_node = device.device_node

        if device_node is not None:
            return device_node

        error("Unable to find a touch device.")
        sys.exit(1)

    def handle_btn_touch(self, event):
        if event.value != 0:
            t = Touch(tv2us(event.sec, event.usec))
            self.touches.append(t)
        else:
            self.touches[-1].up = tv2us(event.sec, event.usec)
            msg("\rTouch sequences detected: {}".format(len(self.touches)), end="")

    def handle_key(self, event):
        tapcodes = [
            libevdev.EV_KEY.BTN_TOOL_DOUBLETAP,
            libevdev.EV_KEY.BTN_TOOL_TRIPLETAP,
            libevdev.EV_KEY.BTN_TOOL_QUADTAP,
            libevdev.EV_KEY.BTN_TOOL_QUINTTAP,
        ]
        if event.code in tapcodes and event.value > 0:
            if not self.warned:
                self.warned = True
                error(
                    "\rThis tool cannot handle multiple fingers, output will be invalid"
                )
                return

        if event.matches(libevdev.EV_KEY.BTN_TOUCH):
            self.handle_btn_touch(event)

    def handle_syn(self, event):
        if self.touch.dirty:
            self.current_sequence().append(self.touch)
            self.touch = Touch(
                major=self.touch.major,
                minor=self.touch.minor,
                orientation=self.touch.orientation,
            )

    def handle_event(self, event):
        if event.matches(libevdev.EV_KEY):
            self.handle_key(event)

    def read_events(self):
        while True:
            for event in self.events():
                self.handle_event(event)

    def print_summary(self):
        deltas = sorted(t.tdelta for t in self.touches)

        dmax = max(deltas)
        dmin = min(deltas)

        ndeltas = len(deltas)

        davg = sum(deltas) / ndeltas
        dmedian = deltas[int(ndeltas / 2)]
        d95pc = deltas[int(ndeltas * 0.95)]
        d90pc = deltas[int(ndeltas * 0.90)]

        print("Time: ")
        print("  Max delta: {}ms".format(int(dmax)))
        print("  Min delta: {}ms".format(int(dmin)))
        print("  Average delta: {}ms".format(int(davg)))
        print("  Median delta: {}ms".format(int(dmedian)))
        print("  90th percentile: {}ms".format(int(d90pc)))
        print("  95th percentile: {}ms".format(int(d95pc)))

    def print_dat(self):
        print("# libinput-measure-touchpad-tap")
        print(
            textwrap.dedent(
                """\
              # File contents:
              #    This file contains multiple prints of the data in
              #    different sort order. Row number is index of touch
              #    point within each group. Comparing data across groups
              #    will result in invalid analysis.
              # Columns (1-indexed):
              # Group 1, sorted by time of occurrence
              #  1: touch down time in ms, offset by first event
              #  2: touch up time in ms, offset by first event
              #  3: time delta in ms);
              # Group 2, sorted by touch down-up delta time (ascending)
              #  4: touch down time in ms, offset by first event
              #  5: touch up time in ms, offset by first event
              #  6: time delta in ms
              """
            )
        )

        deltas = [t for t in self.touches]
        deltas_sorted = sorted(deltas, key=lambda t: t.tdelta)

        offset = deltas[0].down

        for t1, t2 in zip(deltas, deltas_sorted):
            print(
                t1.down - offset,
                t1.up - offset,
                t1.tdelta,
                t2.down - offset,
                t2.up - offset,
                t2.tdelta,
            )

    def print(self, format):
        if not self.touches:
            error("No tap data available")
            return

        if format == "summary":
            self.print_summary()
        elif format == "dat":
            self.print_dat()


def main(args):
    parser = argparse.ArgumentParser(
        description="Measure tap-to-click properties of devices"
    )
    parser.add_argument(
        "path",
        metavar="/dev/input/event0",
        nargs="?",
        type=str,
        help="Path to device (optional)",
    )
    parser.add_argument(
        "--format",
        metavar="format",
        choices=["summary", "dat"],
        default="summary",
        help='data format to print ("summary" or "dat")',
    )
    args = parser.parse_args()

    if not sys.stdout.isatty():
        global print_dest
        print_dest = sys.stderr

    try:
        device = Device(args.path)
        error(
            "Ready for recording data.\n"
            "Tap the touchpad multiple times with a single finger only.\n"
            "For useful data we recommend at least 20 taps.\n"
            "Ctrl+C to exit"
        )
        device.read_events()
    except KeyboardInterrupt:
        msg("")
        device.print(args.format)
    except (PermissionError, OSError) as e:
        error("Error: failed to open device. {}".format(e))
    except InvalidDeviceError as e:
        error("Error: {}".format(e))


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# -*- coding: utf-8
# vim: set expandtab shiftwidth=4:
# -*- Mode: python; coding: utf-8; indent-tabs-mode: nil -*- */
#
# Copyright © 2024 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the 'Software'),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
#
# Prints the data from a libinput recording in a table format to ease
# debugging.
#
# Input is a libinput record yaml file

from dataclasses import dataclass
import argparse
import os
import sys
import yaml
import libevdev

COLOR_RESET = "\x1b[0m"
COLOR_RED = "\x1b[6;31m"


def micros(e: libevdev.InputEvent):
    return e.usec + e.sec * 1_000_000


@dataclass
class Timestamp:
    sec: int
    usec: int

    @property
    def micros(self) -> int:
        return self.usec + self.sec * 1_000_000


@dataclass
class ButtonFrame:
    delta_ms: int  # delta time to last button (not evdev!) frame
    evdev_delta_ms: int  # delta time to last evdev frame
    events: list[libevdev.InputEvent]  # BTN_ events only

    @property
    def timestamp(self) -> Timestamp:
        e = self.events[0]
        return Timestamp(e.sec, e.usec)

    def value(self, code: libevdev.EventCode) -> bool | None:
        for e in self.events:
            if e.matches(code):
                return e.value
        return None

    def values(self, codes: list[libevdev.EventCode]) -> list[bool | None]:
        return [self.value(code) for code in codes]


def frames(events):
    last_timestamp = None
    current_frame = None
    last_frame = None
    for e in events:
        if last_timestamp is None:
            last_timestamp = micros(e)
        if e.type == libevdev.EV_SYN:
            last_timestamp = micros(e)
            if current_frame is not None:
                yield current_frame
            last_frame = current_frame
            current_frame = None
        elif e.type == libevdev.EV_KEY:
            if e.code.name.startswith("BTN_") and not e.code.name.startswith(
                "BTN_TOOL_"
            ):
                timestamp = micros(e)
                evdev_delta = (timestamp - last_timestamp) // 1000

                if last_frame is not None:
                    delta = (timestamp - last_frame.timestamp.micros) // 1000
                else:
                    delta = 0

                if current_frame is None:
                    current_frame = ButtonFrame(
                        delta_ms=delta, evdev_delta_ms=evdev_delta, events=[e]
                    )
                else:
                    current_frame.events.append(e)


def main(argv):
    parser = argparse.ArgumentParser(description="Display button events in a recording")
    parser.add_argument(
        "--threshold",
        type=int,
        default=25,
        help="Mark any time delta above this threshold (in ms)",
    )
    parser.add_argument(
        "path", metavar="recording", nargs=1, help="Path to libinput-record YAML file"
    )
    args = parser.parse_args()
    isatty = os.isatty(sys.stdout.fileno())
    if not isatty:
        global COLOR_RESET
        global COLOR_RED
        COLOR_RESET = ""
        COLOR_RED = ""

    yml = yaml.safe_load(open(args.path[0]))
    if yml["ndevices"] > 1:
        print(f"WARNING: Using only first {yml['ndevices']} devices in recording")
    device = yml["devices"][0]
    if not device["events"]:
        print("No events found in recording")
        sys.exit(1)

    def events():
        """
        Yields the next event in the recording
        """
        for event in device["events"]:
            for evdev in event.get("evdev", []):
                yield libevdev.InputEvent(
                    code=libevdev.evbit(evdev[2], evdev[3]),
                    value=evdev[4],
                    sec=evdev[0],
                    usec=evdev[1],
                )

    # These are the buttons we possibly care about, but we filter to the ones
    # found on this device anyway
    buttons = [
        libevdev.EV_KEY.BTN_LEFT,
        libevdev.EV_KEY.BTN_MIDDLE,
        libevdev.EV_KEY.BTN_RIGHT,
        libevdev.EV_KEY.BTN_SIDE,
        libevdev.EV_KEY.BTN_EXTRA,
        libevdev.EV_KEY.BTN_FORWARD,
        libevdev.EV_KEY.BTN_BACK,
        libevdev.EV_KEY.BTN_TASK,
        libevdev.EV_KEY.BTN_TOUCH,
        libevdev.EV_KEY.BTN_STYLUS,
        libevdev.EV_KEY.BTN_STYLUS2,
        libevdev.EV_KEY.BTN_STYLUS3,
        libevdev.EV_KEY.BTN_0,
        libevdev.EV_KEY.BTN_1,
        libevdev.EV_KEY.BTN_2,
        libevdev.EV_KEY.BTN_3,
        libevdev.EV_KEY.BTN_4,
        libevdev.EV_KEY.BTN_5,
        libevdev.EV_KEY.BTN_6,
        libevdev.EV_KEY.BTN_7,
        libevdev.EV_KEY.BTN_8,
        libevdev.EV_KEY.BTN_9,
    ]

    def filter_buttons(buttons):
        return filter(
            lambda c: c in buttons,
            map(lambda c: libevdev.evbit("EV_KEY", c), device["evdev"]["codes"][1]),
        )

    buttons = list(filter_buttons(buttons))

    # all BTN_STYLUS will have a header of S - meh
    btn_headers = " │ ".join(b.name[4] for b in buttons)
    print(f"{'Timestamp':^13s} │ {'Delta':^8s} │ {btn_headers}")
    last_btn_vals = [None] * len(buttons)

    def btnchar(b, last):
        if b == 1:
            return "┬"
        if b == 0:
            return "┴"
        return "│" if last else " "

    for frame in frames(events()):
        ts = frame.timestamp
        if frame.timestamp.micros > 0 and frame.delta_ms < args.threshold:
            color = COLOR_RED
        else:
            color = ""
        btn_vals = frame.values(buttons)
        btn_strs = " │ ".join(
            [btnchar(b, last) for b, last in zip(btn_vals, last_btn_vals)]
        )

        last_btn_vals = [
            b if b is not None else last for b, last in zip(btn_vals, last_btn_vals)
        ]

        print(
            f"{color}{ts.sec:6d}.{ts.usec:06d} │ {frame.delta_ms:6d}ms │ {btn_strs}{COLOR_RESET}"
        )


if __name__ == "__main__":
    try:
        main(sys.argv)
    except BrokenPipeError:
        pass
